
import datetime as dt
from calendar import day_abbr, month_abbr
from colorsys import hsv_to_rgb
from logging import getLogger

import numpy as np
import pandas as pd
from bokeh.io import show
from bokeh.models import Rect, ColumnDataSource, Circle, Arc, MultiLine, SaveTool, HoverTool, Label, Plot, Range1d, \
    Title
from bokeh.palettes import *
from math import pi

from .utils import tooltip
from ...lib.data import linscale
from ...lib.date import time_to_local_time, YMD, to_time
from ...stoats.names import LOCAL_TIME, DIRECTION, ASPECT_RATIO, ACTIVE_DISTANCE, TOTAL_CLIMB, like, _d, FITNESS_D_ANY, \
    GROUP

log = getLogger(__name__)

_max, _min = max, min

CALENDAR = 'Calendar'
X, Y, XS, YS = 'X', 'Y', 'Xs', 'Ys'
ALPHA, ANGLE, ARC, COLOR, END, FILL, RADIUS, SIZE, START, WEEK, WIDTH = \
    'Alpha', 'Angle', 'Arc', 'Color', 'End', 'Fill', 'Radius', 'Size', 'Start', 'Week', 'Width'

def _cal(name): return f'{CALENDAR} {name}'

CALENDAR_X = _cal(X.upper())
CALENDAR_Y = _cal(Y.upper())
CALENDAR_XS = _cal(XS)
CALENDAR_YS = _cal(YS)
CALENDAR_ALPHA = _cal(ALPHA)
CALENDAR_ARC = _cal(ARC)
CALENDAR_ARC_RADIUS = _cal(f'{ARC} {RADIUS}')
CALENDAR_RADIUS = _cal(RADIUS)
CALENDAR_SIZE = _cal(SIZE)
CALENDAR_WEEK = _cal(WEEK)
CALENDAR_ANGLE = _cal(ANGLE)
CALENDAR_WIDTH = _cal(WIDTH)
CALENDAR_START = _cal(START)
CALENDAR_END = _cal(END)
CALENDAR_COLOR = _cal(COLOR)
CALENDAR_FILL = _cal(FILL)


K2R = [f'#{i:02x}0000'.upper() for i in range(256)]
B2R = [f'#{i:02x}00{0xff-i:02x}'.upper() for i in range(256)]
K2W = [f'#{i:02x}{i:02x}{i:02x}'.upper() for i in range(256)]


class Calendar:

    '''
    The best way to understand this is probably to see the calendar template which shows three examples.
    Or read the std_ plot definitions.

    Data are stored in the dataframe, using column names prefixed by 'Calendar'.  These can then be referred
    to when plotting (or constant values can be provided).
    '''

    def __init__(self, df, scale=13, border_day=0.25, border_month=0.4, border_year=0.4, title=None,
                 hover=None, not_hover=tuple(), all_hover=False):

        self._hover_args = (hover, not_hover, all_hover)

        start, finish = df.index.min(), df.index.max()
        delta_year = 7 * (1 + border_day) + border_year

        self._df = df
        self._df_all = pd.DataFrame(data={CALENDAR_SIZE: 1, CALENDAR_RADIUS: 0.5},
                                    index=pd.date_range(start, finish))

        self._set_date(self._df)
        self._set_xy(self._df, start, finish, border_day, border_month, delta_year)
        self._set_date(self._df_all)
        self._set_xy(self._df_all, start, finish, border_day, border_month, delta_year)

        xlo = self._df_all[CALENDAR_X].min()
        xhi = self._df_all[CALENDAR_X].max()
        ylo = self._df_all[CALENDAR_Y].min()
        yhi = self._df_all[CALENDAR_Y].max()
        nx = int(scale * (xhi - xlo))
        ny = int(scale * (yhi - ylo))

        plot = Plot(x_range=Range1d(xlo - 4, xhi + 4), y_range=Range1d(ylo - 1, yhi + 4),
                    width=nx, height=ny, title=Title(text=title))
        self._add_labels(plot, start, finish, xlo, xhi, yhi, border_day, delta_year)
        self._plot = plot

    @staticmethod
    def _set_date(df):
        df.loc[:, LOCAL_TIME] = df.index.map(lambda x: time_to_local_time(x.to_pydatetime(), YMD))

    @staticmethod
    def _set_xy(df, start, finish, border_day, border_month, delta_year):
        offset = dict((y, dt.date(year=y, month=1, day=1).weekday() - 1)
                      for y in range(start.year, finish.year+1))
        df.loc[:, CALENDAR_WEEK] = df.index.map(lambda index: (index.dayofyear + offset[index.year]) // 7)
        df.loc[:, CALENDAR_X] = df[CALENDAR_WEEK] * (1 + border_day) + df.index.month * border_month
        df.loc[:, CALENDAR_Y] = -1 * (df.index.dayofweek * (1 + border_day) +
                                      (df.index.year - start.year) * delta_year)

    @staticmethod
    def _add_labels(plot, start, finish, xlo, xhi, yhi, border_day, delta_year):
        for y in range(start.year, finish.year+1):
            plot.add_layout(Label(x=xhi + 1.1, y=(start.year - y - 0.4) * delta_year, text=str(y),
                                        text_align='center', text_baseline='bottom', angle=3*pi/2))
            for d, text in enumerate(day_abbr):
                plot.add_layout(Label(x=xlo - 1.5, y=(start.year - y) * delta_year - d * (1 + border_day),
                                      text=text, text_align='right', text_baseline='middle',
                                      text_font_size='7pt', text_color='grey'))
        dx = (xhi - xlo) / 12
        for m, text in enumerate(month_abbr[1:]):
            plot.add_layout(Label(x=xlo + dx * (m + 0.5), y=yhi + 1.5,
                                  text=text, text_align='center', text_baseline='bottom'))
        plot.toolbar.logo = None

    @staticmethod
    def _build_tools(df, hover, not_hover, all_hover):
        tools = [SaveTool()]
        if hover is None:
            hover = [col for col in df.columns if all_hover or (not col.startswith(CALENDAR) and not col in not_hover)]
        if hover:
            # names ensures that background has no hover
            tools.append(HoverTool(tooltips=[tooltip(col) for col in hover], names=['with_hover']))
        return tools

    def set_constant(self, dest, value):
        self._df.loc[:, dest] = value

    def set_linear(self, dest, name, lo=None, hi=None, min=0, max=1, gamma=1):
        self._df.loc[:, dest] = linscale(self._df[name], lo=lo, hi=hi, min=min, max=max, gamma=gamma)

    def set_size(self, name, lo=None, hi=None, min=0, max=1, gamma=1):
        self.set_linear(CALENDAR_SIZE, name, lo=lo, hi=hi, min=min, max=max, gamma=gamma)
        self._df.loc[:, CALENDAR_RADIUS] = self._df[CALENDAR_SIZE] / 2

    def set_palette(self, name, palette, lo=None, hi=None, min=0, max=1, gamma=1, nan='white'):
        n = len(palette)
        self._df.loc[:, CALENDAR_COLOR] = linscale(self._df[name], lo=lo, hi=hi, min=min, max=max, gamma=gamma). \
            map(lambda x: nan if np.isnan(x) else palette[_max(0, _min(n-1, int(x * n)))])

    def set_arc(self, angle, width, size=None, delta_radius=0, lo=0, hi=2, min=30, max=180, gamma=1):
        angle = 90 - self._df[angle]  # +ve anticlock from x
        width = linscale(self._df[width], lo=lo, hi=hi, min=min, max=max, gamma=gamma)
        self._df.loc[:, CALENDAR_START] = pi * (angle - width/2) / 180
        self._df.loc[:, CALENDAR_END] = pi * (angle + width/2) / 180
        if size is None:
            self._df.loc[:, CALENDAR_ARC_RADIUS] = self._df[CALENDAR_RADIUS]
        else:
            self._df.loc[:, CALENDAR_ARC_RADIUS] = size / 2
        self._df.loc[:, CALENDAR_ARC_RADIUS] += delta_radius

    def std_distance(self):
        self.background('square', fill_alpha=0, line_alpha=1, color='lightgrey')
        self.set_size(ACTIVE_DISTANCE, min=0.1)
        self.foreground('square', fill_alpha=1, line_alpha=0, color='black')
        self.show()

    def std_distance_climb_direction(self):
        self.background('circle', fill_alpha=1, line_alpha=0, color='#F0F0F0')
        self.set_size(ACTIVE_DISTANCE, min=0.2, max=1.1)
        self.set_palette(TOTAL_CLIMB, K2R, gamma=0.3)  # more red
        self.foreground('circle', fill_alpha=1, line_alpha=0)
        self.foreground('circle', fill_alpha=0, line_alpha=1, color='grey')
        self.set_arc(DIRECTION, ASPECT_RATIO, delta_radius=0.2)
        self.foreground('arc', fill_alpha=0, line_alpha=1, color='black')
        self.show()

    def std_distance_fitness_direction(self):
        fitness = sorted(like(_d(FITNESS_D_ANY), self._df.columns))[0]
        self.background('circle', fill_alpha=0, line_alpha=1, color='lightgrey')
        self.set_size(ACTIVE_DISTANCE, min=0.2, max=1.1)
        self.set_palette(fitness, B2R, gamma=0.7)
        self.foreground('circle', fill_alpha=1, line_alpha=0)
        self.set_arc(DIRECTION, ASPECT_RATIO, delta_radius=0.2)
        self.foreground('arc', fill_alpha=0, line_alpha=1)
        self.show()

    def std_group_distance_climb_direction(self):

        n = int(self._df[GROUP].max()) + 1

        def pastel(i):
            r, g, b = [int(x * 255) for x in hsv_to_rgb((i % n) / n, 0.2, 1)]
            return f'#{r:02x}{g:02x}{b:02x}'.upper()
        palette = list(pastel(7 * i) for i in range(n))

        self.background('square', fill_alpha=0, line_alpha=1, color='#F0F0F0')
        self.set_palette(GROUP, palette)
        self.set_constant(CALENDAR_SIZE, 1)
        self.foreground('square', fill_alpha=1, line_alpha=0)
        self.set_size(ACTIVE_DISTANCE, min=0.2, max=0.8)
        self.set_palette(TOTAL_CLIMB, K2W, gamma=0.3)
        self.foreground('circle', fill_alpha=1, line_alpha=0)
        self.foreground('circle', fill_alpha=0, line_alpha=1, color='black')
        self.set_arc(DIRECTION, ASPECT_RATIO, delta_radius=0.2)
        self.foreground('arc', fill_alpha=0, line_alpha=1, color='black')
        self.show()

    def show(self):
        self._plot.add_tools(*self._build_tools(self._df, *self._hover_args))
        show(self._plot)

    def background(self, glyph, fill_alpha=CALENDAR_FILL, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        if glyph == 'square':
            self._rect(self._df_all, fill_alpha=fill_alpha, line_alpha=line_alpha, color=color)
        elif glyph == 'circle':
            self._circle(self._df_all, fill_alpha=fill_alpha, line_alpha=line_alpha, color=color)
        else:
            raise Exception(f'Unexpected background glyph {glyph}')

    def foreground(self, glyph, fill_alpha=CALENDAR_FILL, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        if glyph == 'square':
            self._rect(self._df, fill_alpha=fill_alpha, line_alpha=line_alpha, color=color)
        elif glyph == 'circle':
            self._circle(self._df, fill_alpha=fill_alpha, line_alpha=line_alpha, color=color)
        elif glyph == 'arc':
            self._arc(self._df, line_alpha=line_alpha, color=color)
        elif glyph == 'sqarc':
            self._sqarc(self._df, line_alpha=line_alpha, color=color)
        else:
            raise Exception(f'Unexpected foreground glyph {glyph}')

    def _rect(self, df, fill_alpha=CALENDAR_FILL, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        rect = Rect(x=CALENDAR_X, y=CALENDAR_Y, width=CALENDAR_SIZE, height=CALENDAR_SIZE,
                    fill_color=color, fill_alpha=fill_alpha,
                    line_color=color, line_alpha=line_alpha)
        return self._plot.add_glyph(ColumnDataSource(df), rect, name='with_hover' if df is self._df else None)

    def _circle(self, df, fill_alpha=CALENDAR_FILL, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        circle = Circle(x=CALENDAR_X, y=CALENDAR_Y, radius=CALENDAR_RADIUS,
                        fill_color=color, fill_alpha=fill_alpha,
                        line_color=color, line_alpha=line_alpha)
        return self._plot.add_glyph(ColumnDataSource(df), circle, name='with_hover' if df is self._df else None)

    def _arc(self, df, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        arc = Arc(x=CALENDAR_X, y=CALENDAR_Y, radius=CALENDAR_ARC_RADIUS,
                  start_angle=CALENDAR_START, end_angle=CALENDAR_END, direction='anticlock',
                  line_color=color, line_alpha=line_alpha)
        return self._plot.add_glyph(ColumnDataSource(df), arc, name='with_hover' if df is self._df else None)

    def _sqarc(self, df, line_alpha=CALENDAR_ALPHA, color=CALENDAR_COLOR):
        df = df.dropna().copy()  # copy to avoid set on view errors
        xys = list(_corners(df))
        xs, ys = list(zip(*xys))
        df.loc[:, CALENDAR_XS] = xs
        df.loc[:, CALENDAR_YS] = ys
        multi = MultiLine(xs=CALENDAR_XS, ys=CALENDAR_YS, line_color=color, line_alpha=line_alpha)
        self._plot.add_glyph(ColumnDataSource(df), multi)


# helpers for generating square arc multiline (a lot of work for not much result)

def _edge(linear, x, y, r):
    if np.isnan(linear): return x, y
    d = 2 * r * (linear - int(linear))
    if linear < 1: return x + r - d, y + r
    if linear < 2: return x - r, y + r - d
    if linear < 3: return x - r + d, y - r
    return x + r, y - r + d


def _corner(linear, x, y, r):
    if np.isnan(linear): return x, y
    if linear < 1: return x + r, y + r
    if linear < 2: return x - r, y + r
    if linear < 3: return x - r, y - r
    return x + r, y - r


def _line_corners(x, y, r, start, end):
    yield _edge(start, x, y, r)
    while start < end:
        next_corner = int(start + 1)
        if next_corner < end:
            start = next_corner
            yield _corner(start, x, y, r)
        else:
            start = end
            yield _edge(start, x, y, r)


def _linearize(angle):
    linear = (angle - pi/4) / (pi/2)
    while linear <= 0: linear += 4
    while linear > 4: linear -= 4
    return linear


def _corners(df):
    for _, row in df.iterrows():
        x, y, r = row[CALENDAR_X], row[CALENDAR_Y], row[CALENDAR_ARC_RADIUS]
        start, end = _linearize(row[CALENDAR_START]), _linearize(row[CALENDAR_END])
        xys = list(_line_corners(x, y, r, start, end))
        yield list(zip(*xys))




import asyncio
from _signal import pause
from logging import getLogger
from os import makedirs
from threading import Thread, Event
from time import sleep

from notebook.notebookapp import NotebookApp
from tornado.platform.asyncio import AnyThreadEventLoopPolicy

from ..commands.args import NOTEBOOKS, JUPYTER, SERVICE, VERBOSITY
from ..lib.workers import command_root
from ..squeal import SystemConstant, SystemProcess

log = getLogger(__name__)


class JupyterServer(NotebookApp):

    def __init__(self, started, **kwargs):
        self._started = started
        super().__init__(**kwargs)

    @property
    def log_level(self):
        # this silences jupyter's logging
        # the existing mechanism only lets you set a value of 50, which means that "critical" messages
        # are still logged, and the decidedly non-critical usage instructions are printed.
        return 60

    def init_signal(self):
        log.debug('Skipping signal init')

    def start(self):
        self._started.set()
        super().start()


class JupyterController:

    def __init__(self, db, args, max_retries=5, retry_secs=3):
        self._db = db
        self._notebooks = args.path(NOTEBOOKS)
        self._log_level = args[VERBOSITY]
        self._max_retries = max_retries
        self._retry_secs = retry_secs

    def start_service(self, restart=False):
        with self._db.session_context() as s:
            if SystemProcess.exists_any(s, JupyterServer):
                log.debug('Jupyter already running')
                if restart:
                    self.stop_service()
                else:
                    return
            log.debug('Starting remote Jupyter server')
            ch2 = command_root()
            log_name = 'jupyter-service.log'
            cmd = f'{ch2} -v{self._log_level} -l {log_name} --{NOTEBOOKS} {self._notebooks} {JUPYTER} {SERVICE}'
            SystemProcess.run(s, cmd, log_name, JupyterServer)
            retries = 0
            while not SystemProcess.exists_any(s, JupyterServer):
                retries += 1
                if retries > self._max_retries:
                    raise Exception('Jupyter server did not start')
                sleep(self._retry_secs)
            log.info('Jupyter server started')

    def stop_service(self):
        log.info('Stopping any running Jupyter server')
        with self._db.session_context() as s:
            SystemProcess.delete_all(s, JupyterServer)
            SystemConstant.delete(s, SystemConstant.JUPYTER_URL)
            SystemConstant.delete(s, SystemConstant.JUPYTER_DIR)

    def connection_url(self):
        self.start_service()
        with self._db.session_context() as s:
            return SystemConstant.get(s, SystemConstant.JUPYTER_URL)

    def notebook_dir(self):
        self.start_service()
        with self._db.session_context() as s:
            return SystemConstant.get(s, SystemConstant.JUPYTER_DIR)

    def run_local(self):
        self.stop_service()

        log.info('Starting a local Jupyter server')
        log.debug(f'Creating {self._notebooks}')
        makedirs(self._notebooks, exist_ok=True)

        started = Event()

        def start():
            log.info('Starting Jupyter server in separate thread')
            asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
            JupyterServer.launch_instance(['--notebook-dir', self._notebooks], started=started)

        t = Thread(target=start)
        t.daemon = True
        t.start()
        started.wait()  # set in JupyterServer.start() which is as late as we can get in startup
        log.debug('Separate thread started')

        while not hasattr(JupyterServer._instance, 'connection_url') or not JupyterServer._instance.connection_url:
            log.debug('Waiting for connection URL')
            sleep(1)

        with self._db.session_context() as s:
            SystemConstant.set(s, SystemConstant.JUPYTER_URL, JupyterServer._instance.connection_url, force=True)
            SystemConstant.set(s, SystemConstant.JUPYTER_DIR, self._notebooks, force=True)

        log.info('Jupyter server started')
        pause()


__CONTROLLER_SINGLETON = None


def start_controller(db, args):
    global __CONTROLLER_SINGLETON
    if __CONTROLLER_SINGLETON:
        raise Exception('Jupyter controller already started')
    __CONTROLLER_SINGLETON = JupyterController(db, args)


def get_controller():
    if not __CONTROLLER_SINGLETON:
        raise Exception('Jupyter controller not started')
    return __CONTROLLER_SINGLETON




import datetime as dt
from json import dumps, loads
from logging import getLogger
from pydoc import locate

from sqlalchemy import TypeDecorator, Integer, Float, Text

from ..lib.date import to_time, to_date
from ..lib.schedule import Schedule

log = getLogger(__name__)


class Date(TypeDecorator):

    impl = Integer

    def process_literal_param(self, date, dialect):
        if date is None:
            return date
        return to_date(date).toordinal()

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        if value < 1:  # bootstrap TopicJournal.date where we set values to 0
            return None
        else:
            return dt.date.fromordinal(value)


class Time(TypeDecorator):

    impl = Float

    def process_literal_param(self, time, dialect):
        if time is None:
            return time
        else:
            converted = to_time(time).timestamp()
            # log.debug(f'Time (writing): converted {time} to {converted}')
            return converted

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            converted = dt.datetime.fromtimestamp(value, dt.timezone.utc)
            # log.debug(f'Time (reading): converted {value} to {converted}')
            return converted


CLS_CACHE = {}


class Cls(TypeDecorator):

    impl = Text

    def process_literal_param(self, cls, dialect):
        return long_cls(cls)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return lookup_cls(value)


def long_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__module__ + '.' + cls.__name__
    return cls


def lookup_cls(value):
    # https://stackoverflow.com/a/24815361
    if not value:
        return None
    if value not in CLS_CACHE:
        CLS_CACHE[value] = locate(value)
    if not CLS_CACHE[value]:
        raise Exception('Cannot find %s' % value)
    return CLS_CACHE[value]


class ShortCls(TypeDecorator):

    impl = Text

    def process_literal_param(self, cls, dialect):
        return short_cls(cls)

    process_bind_param = process_literal_param


def short_cls(cls):
    if cls is None:
        return cls
    if not isinstance(cls, str) and not isinstance(cls, type):
        cls = type(cls)
    if isinstance(cls, type):
        cls = cls.__name__
    return cls


class Str(TypeDecorator):

    impl = Text

    def process_literal_param(self, value, dialect):
        if value is None:
            return value
        if isinstance(value, int):
            raise Exception('Passing primary key instead of class?')
        return str(value)

    process_bind_param = process_literal_param


class NullStr(TypeDecorator):
    '''
    None (NULL) values are converted to 'None'.

    Use for constraint, where NULLs are distinct in UNIQUE constraints.
    '''

    impl = Text

    def process_literal_param(self, value, dialect):
        return str(value)

    process_bind_param = process_literal_param

    coerce_to_is_types = tuple()  # don't want "IS NULL" when comparing with None


class Json(TypeDecorator):

    impl = Text

    def process_literal_param(self, value, dialect):
        return dumps(value)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        return loads(value)


class Sched(TypeDecorator):

    impl = Text

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        return str(sched)

    process_bind_param = process_literal_param

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return Schedule(value)


class OpenSched(Sched):

    def process_literal_param(self, sched, dialect):
        if sched is None:
            return sched
        if not isinstance(sched, Schedule):
            sched = Schedule(sched)
        sched.start = None
        sched.finish = None
        return str(sched)

    process_bind_param = process_literal_param


class Sort(TypeDecorator):

    impl = Integer

    def process_literal_param(self, value, dialect):
        if callable(value):
            value = value()
        return value

    process_bind_param = process_literal_param




from hashlib import md5
from logging import getLogger
from os import stat
from shutil import get_terminal_size
from time import time

from sqlalchemy import desc

from .date import to_time
from ch2.squeal.utils import add
from ..squeal.tables.fit import FileScan

log = getLogger(__name__)


def terminal_width(width=None):
    return get_terminal_size()[0] if width is None else width


def tui(command):
    def wrapper(*args, **kargs):
        return command(*args, **kargs)
    wrapper.tui = True
    wrapper.__doc__ = command.__doc__
    return wrapper


# https://stackoverflow.com/a/3431838
def md5_hash(file_path):
    hash = md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash.update(chunk)
    return hash.hexdigest()


def for_modified_files(log, s, paths, callback, owner, force=False):
    '''
    This takes a callback because we need to know whether to mark the file as read or not
    after processing.  The callback should return True on success.

    The session is used throughout, but not passed to the callback.  The callback can
    contain the same session as internal state or create its own.  We avoid open
    transactions across the callback.
    '''

    for file_path in paths:

        last_modified = to_time(stat(file_path).st_mtime)
        hash = md5_hash(file_path)
        path_scan = s.query(FileScan). \
            filter(FileScan.path == file_path,
                   FileScan.owner == owner).one_or_none()

        # get last scan and make sure it's up-to-date
        if path_scan:
            if hash != path_scan.md5_hash:
                log.warning('File at %s appears to have changed since last read on %s')
                path_scan.md5_hash = hash
                path_scan.last_scan = 0.0
        else:
            # need to_time here because it's not roundtripped via the database to convert for use below
            path_scan = add(s, FileScan(path=file_path, owner=owner,
                                        md5_hash=hash, last_scan=to_time(0.0)))
            s.flush()  # want this to appear in queries below

        # only look at hash if we are going to process anyway
        if force or last_modified > path_scan.last_scan:

            hash_scan = s.query(FileScan). \
                filter(FileScan.md5_hash == hash,
                       FileScan.owner == owner).\
                order_by(desc(FileScan.last_scan)).limit(1).one()  # must exist as path_scan is a candidate
            if hash_scan.path != path_scan.path:
                log.warning('Ignoring duplicate file (details in debug log)')
                log.debug('%s' % file_path)
                log.debug('%s' % hash_scan.path)
                # update the path to avoid triggering in future
                path_scan.last_scan = hash_scan.last_scan

            s.commit()

            if force or last_modified > hash_scan.last_scan:
                if callback(file_path):
                    log.debug('Marking %s as scanned' % file_path)
                    path_scan.last_scan = last_modified  # maybe use now?
                    s.commit()
                else:
                    log.debug('Not marking %s as scanned' % file_path)


def filter_modified_files(s, paths, owner, force=False):

    modified = []

    for file_path in paths:

        last_modified = to_time(stat(file_path).st_mtime)
        hash = md5_hash(file_path)
        path_scan = s.query(FileScan). \
            filter(FileScan.path == file_path,
                   FileScan.owner == owner).one_or_none()

        # get last scan and make sure it's up-to-date
        if path_scan:
            if hash != path_scan.md5_hash:
                log.warning('File at %s appears to have changed since last read on %s')
                path_scan.md5_hash = hash
                path_scan.last_scan = 0.0
        else:
            # need to_time here because it's not roundtripped via the database to convert for use below
            path_scan = add(s, FileScan(path=file_path, owner=owner,
                                        md5_hash=hash, last_scan=to_time(0.0)))
            s.flush()  # want this to appear in queries below

        # only look at hash if we are going to process anyway
        if force or last_modified > path_scan.last_scan:

            hash_scan = s.query(FileScan). \
                filter(FileScan.md5_hash == hash,
                       FileScan.owner == owner).\
                order_by(desc(FileScan.last_scan)).limit(1).one()  # must exist as path_scan is a candidate
            if hash_scan.path != path_scan.path:
                log.warning('Ignoring duplicate file (details in debug log)')
                log.debug('%s' % file_path)
                log.debug('%s' % hash_scan.path)
                # update the path to avoid triggering in future
                path_scan.last_scan = hash_scan.last_scan

            if force or last_modified > hash_scan.last_scan:
                modified.append(file_path)

    s.commit()
    return modified


def update_scan(s, file_path, owner):
    path_scan = s.query(FileScan). \
        filter(FileScan.path == file_path,
               FileScan.owner == owner).one()
    path_scan.last_scan = time()




from bokeh.plotting import output_notebook, show

from ch2.data import *
from ch2.lib import *
from ch2.squeal import *
from ch2.uranus.decorator import template


@template
def all_activities(start: to_date, finish: to_date):

    f'''
    # All Activities: {start.split()[0]} - {finish.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Build Maps
    
    Loop over activities, retrieve data, and construct maps. 
    '''

    s = session('-v2')
    maps = [map_thumbnail(100, 120, data.resample('1min').mean())
            for data in (activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                             activity_journal=aj)
                         for aj in s.query(ActivityJournal).
                             filter(ActivityJournal.start >= local_date_to_time(start),
                                    ActivityJournal.start < local_date_to_time(finish)).all())
            if len(data.dropna()) > 10]
    print(f'Found {len(maps)} activities')

    '''
    ## Display Maps
    '''

    output_notebook()
    show(tile(maps, 8))




import datetime as dt
from collections import defaultdict, Counter
from collections.abc import Mapping, Sequence
from logging import getLogger

import numpy as np
import pandas as pd
from sqlalchemy import inspect, select, and_, or_, distinct
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import coalesce

from .coasting import CoastingBookmark
from ..lib.data import kargs_to_attr
from ..lib.date import local_time_to_time, time_to_local_time, YMD, HMS
from ..squeal import StatisticName, StatisticJournal, StatisticJournalInteger, ActivityJournal, \
    StatisticJournalFloat, StatisticJournalText, Interval, StatisticMeasure, Source
from ..squeal.database import connect, ActivityTimespan, ActivityGroup, ActivityBookmark, StatisticJournalType, \
    Composite, CompositeComponent, ActivityNearby
from ..stoats.display.nearby import nearby_any_time
from ..stoats.names import DELTA_TIME, HEART_RATE, _src, FITNESS_D_ANY, FATIGUE_D_ANY, like, _log, HEART_RATE_BPM, \
    MED_HEART_RATE_BPM
from ..stoats.names import DISTANCE_KM, SPEED_KMH, MED_SPEED_KMH, MED_HR_IMPULSE_10, MED_CADENCE, \
    ELEVATION_M, CLIMB_MS, ACTIVE_TIME_H, ACTIVE_DISTANCE_KM, MED_POWER_ESTIMATE_W, \
    TIMESPAN_ID, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, MED_WINDOW, \
    ELEVATION, SPEED, HR_ZONE, HR_IMPULSE_10, ALTITUDE, CADENCE, TIME, LOCAL_TIME, REST_HR, \
    DAILY_STEPS, ACTIVE_TIME, ACTIVE_DISTANCE, POWER_ESTIMATE, INDEX, GROUP

log = getLogger(__name__)


# in general these functions should (or are being written to) take both ORM objects and parameters
# to retrieve those objects if missing.  so, for example, activity_statistics can be called with
# StatisticName and ActivityJournal instances, or it can be called with more low-level parameters.
# the low-level parameters are often useful interactively but make simplifying asumptions; the ORM
# instances give complete control.


def df(query):
    # https://stackoverflow.com/questions/29525808/sqlalchemy-orm-conversion-to-pandas-dataframe
    return pd.read_sql(query.statement, query.session.bind)


def session(*args):
    ns, db = connect(args)
    return db.session()


def _add_constraint(q, attribute, value, key):
    if value is not None:
        if isinstance(value, Mapping):
            if key in value:
                q = q.filter(attribute == value[key])
        elif isinstance(value, str):
            q = q.filter(attribute == value)
        elif isinstance(value, Sequence):
            q = q.filter(attribute.in_(value))
        else:
            q = q.filter(attribute == value)
    return q


def _collect_statistics(s, names, owner=None, constraint=None):
    if not names:
        names = ['%']
    statistic_ids, statistic_names = set(), set()
    for name in names:
        q = s.query(StatisticName).filter(StatisticName.name.like(name))
        q = _add_constraint(q, StatisticName.owner, owner, name)
        q = _add_constraint(q, StatisticName.constraint, constraint, name)
        for statistic_name in q.all():
            statistic_ids.add(statistic_name.id)
            statistic_names.add(statistic_name.name)
    return statistic_names, statistic_ids


def _tables():
    return kargs_to_attr(sj=inspect(StatisticJournal).local_table,
                         sn=inspect(StatisticName).local_table,
                         sji=inspect(StatisticJournalInteger).local_table,
                         sjf=inspect(StatisticJournalFloat).local_table,
                         sjt=inspect(StatisticJournalText).local_table,
                         inv=inspect(Interval).local_table,
                         at=inspect(ActivityTimespan).local_table,
                         cmp=inspect(Composite).local_table,
                         cc=inspect(CompositeComponent).local_table,
                         src=inspect(Source).local_table)


def _build_statistic_journal_query(statistic_ids, start, finish, source_ids, schedule):

    # use more efficient expression interface and exploit the fact that
    # alternative journal types will be null in an outer join.

    t = _tables()

    q = select([t.sn.c.name, t.sj.c.time, coalesce(t.sjf.c.value, t.sji.c.value, t.sjt.c.value)]). \
        select_from(t.sj.join(t.sn).outerjoin(t.sjf).outerjoin(t.sji).outerjoin(t.sjt)). \
        where(t.sn.c.id.in_(statistic_ids))
    if start:
        q = q.where(t.sj.c.time >= start)
    if finish:
        q = q.where(t.sj.c.time <= finish)
    if source_ids is not None:  # avoid testing DataFrame sequences as bools
        # int() to convert numpy types
        q = q.where(t.sj.c.source_id.in_(int(id) for id in source_ids))
    if schedule:
        q = q.join(t.inv).where(t.inv.c.schedule == schedule)
    q = q.order_by(t.sj.c.time)
    return q


class MissingData(Exception): pass


def make_pad(data, times, statistic_names, quiet=False):
    err_cnt = defaultdict(lambda: 0)

    def pad():
        nonlocal data, times
        n = len(times)
        for name in statistic_names:
            if len(data[name]) != n:
                err_cnt[name] += 1
                if err_cnt[name] <= 1 and not quiet:
                    log.warning('Missing %s at %s (single warning)' % (name, times[-1]))
                data[name].append(None)

    return pad


def activity_journal(s, activity_journal=None, local_time=None, time=None, activity_group_name=None):
    '''
    If activity_journal_id is given, it is returned.

    Otherwise, specify one of (local_time, time) and one of (activity_group_name, activity_group_id).
    '''

    if activity_journal:
        if local_time or time or activity_group_name:
            raise Exception('Activity Journal given, so extra activity-related parameters are unused')
    else:
        if local_time:
            time = local_time_to_time(local_time)
        if not time or not activity_group_name:
            raise Exception('Specify activity_journal or time and activity_group_name')
        try:
            # first, if an activity includes that time
            activity_journal = s.query(ActivityJournal). \
                join(ActivityGroup). \
                filter(ActivityJournal.start <= time,
                       ActivityJournal.finish >= time,
                       ActivityGroup.name.ilike(activity_group_name)).one()
        except NoResultFound:
            # second, if anything in the following 24 hours (eg if just date given)
            activity_journal = s.query(ActivityJournal). \
                join(ActivityGroup). \
                filter(ActivityJournal.start > time,
                       ActivityJournal.finish < time + dt.timedelta(days=1),
                       ActivityGroup.name.ilike(activity_group_name)).one()
        log.info(f'Using Activity Journal {activity_journal}')
    return activity_journal

_activity_journal = activity_journal


def activity_statistics(s, *statistics, owner=None, constraint=None, start=None, finish=None,
                        local_time=None, time=None, bookmarks=None, activity_journal=None,
                        activity_group_name=None, with_timespan=False, check=True):

    if bookmarks:
        if start or finish or local_time or time or activity_journal or activity_group_name:
            raise Exception('Cannot use bookmarks with additional activity constraints')
        return pd.concat(_activity_statistics(s, *statistics, owner=owner, constraint=constraint,
                                              start=bookmark.start, finish=bookmark.finish,
                                              activity_journal=bookmark.activity_journal,
                                              with_timespan=with_timespan, check=check)
                         for bookmark in bookmarks)
    else:
        return _activity_statistics(s, *statistics, owner=owner, constraint=constraint, start=start, finish=finish,
                                    local_time=local_time, time=time, activity_journal=activity_journal,
                                    activity_group_name=activity_group_name, with_timespan=with_timespan, check=check)


def _activity_statistics(s, *statistics, owner=None, constraint=None, start=None, finish=None,
                         local_time=None, time=None, activity_journal=None,
                         activity_group_name=None, with_timespan=False, check=True):

    activity_journal = _activity_journal(s, activity_journal=activity_journal, local_time=local_time,
                                         time=time, activity_group_name=activity_group_name)
    if constraint is None:
        constraint = activity_journal.activity_group
    names = _statistic_names(s, *statistics, owner=owner, constraint=constraint, check=check)
    counts = Counter(name.name for name in names)

    t = _tables()
    ttj = _type_to_journal(t)
    labels = [name.name if counts[name.name] == 1 else f'{name.name} ({name.constraint})' for name in names]
    tables = [ttj[name.statistic_journal_type] for name in names]
    time_select = select([distinct(t.sj.c.time).label("time")]).select_from(t.sj). \
        where(and_(t.sj.c.statistic_name_id.in_([n.id for n in names]),
                   t.sj.c.source_id == activity_journal.id))
    if start:
        time_select = time_select.where(t.sj.c.time >= start)
    if finish:
        time_select = time_select.where(t.sj.c.time <= finish)
    time_select = time_select.order_by("time").alias("sub_time")  # order here avoids extra index
    sub_selects = [select([table.c.value, t.sj.c.time]).
                       select_from(t.sj.join(table)).
                       where(and_(t.sj.c.statistic_name_id == name.id,
                                  t.sj.c.source_id == activity_journal.id)).
                       order_by(t.sj.c.time).  # this doesn't affect plan but seems to speed up query
                       alias(f'sub_{name.name}_{name.constraint}')
                   for name, table in zip(names, tables)]
    # don't call this TIME because even though it's moved to index it somehow blocks the later addition
    # of a TIME column (eg when plotting health statistics)
    selects = [time_select.c.time.label(INDEX)] + \
              [sub.c.value.label(label) for sub, label in zip(sub_selects, labels)]
    sources = time_select
    for sub in sub_selects:
        sources = sources.outerjoin(sub, time_select.c.time == sub.c.time)
    if with_timespan:
        selects += [t.at.c.id.label(TIMESPAN_ID)]
        sources = sources.outerjoin(t.at,
                                    and_(t.at.c.start <= time_select.c.time,
                                         t.at.c.finish > time_select.c.time,
                                         t.at.c.activity_journal_id == activity_journal.id))
    sql = select(selects).select_from(sources)
    return pd.read_sql_query(sql=sql, con=s.connection(), index_col=INDEX)


def statistic_quartiles(s, *statistics, start=None, finish=None, owner=None, constraint=None, source_ids=None,
                        schedule=None):

    # todo - rewrite using new approach

    statistic_names, statistic_ids = _collect_statistics(s, statistics, owner, constraint)
    q = s.query(StatisticMeasure). \
        join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
        join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
        join(Source, StatisticJournal.source_id == Source.id). \
        filter(StatisticName.id.in_(statistic_ids)). \
        filter(StatisticMeasure.quartile != None)
    if start:
        q = q.filter(StatisticJournal.time >= start)
    if finish:
        q = q.filter(StatisticJournal.time <= finish)
    if source_ids is not None:
        q = q.filter(StatisticJournal.source_id.in_(source_ids))
    if schedule:
        q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
            filter(Interval.schedule == schedule)
    log.debug(q)

    raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
    for measure in q.all():
        raw_data[measure.source.start][measure.statistic_journal.statistic_name.name][measure.quartile] = \
            measure.statistic_journal.value
    data, times = defaultdict(list), []
    for time in sorted(raw_data.keys()):
        times.append(time)
        sub_data = raw_data[time]
        for statistic in statistic_names:
            if statistic in sub_data:
                data[statistic].append(sub_data[statistic])
            else:
                data[statistic].append(None)

    return pd.DataFrame(data, index=times)


MIN_PERIODS = 1

def std_activity_statistics(s, local_time=None, time=None, activity_journal=None, activity_group_name=None,
                            with_timespan=True):

    stats = activity_statistics(s, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE,
                                ELEVATION, SPEED, HEART_RATE, HR_ZONE, HR_IMPULSE_10, ALTITUDE, CADENCE,
                                POWER_ESTIMATE,
                                local_time=local_time, time=time, activity_journal=activity_journal,
                                activity_group_name=activity_group_name, with_timespan=with_timespan)

    stats[DISTANCE_KM] = stats[DISTANCE]/1000
    stats[HEART_RATE_BPM] = stats[HEART_RATE]
    stats[SPEED_KMH] = stats[SPEED] * 3.6
    stats[MED_SPEED_KMH] = stats[SPEED].rolling(MED_WINDOW, min_periods=MIN_PERIODS).median() * 3.6
    stats[MED_HEART_RATE_BPM] = stats[HEART_RATE_BPM].rolling(MED_WINDOW, min_periods=MIN_PERIODS).median()
    stats[MED_HR_IMPULSE_10] = stats[HR_IMPULSE_10].rolling(MED_WINDOW, min_periods=MIN_PERIODS).median()
    stats[MED_CADENCE] = stats[CADENCE].rolling(MED_WINDOW, min_periods=MIN_PERIODS).median()
    if POWER_ESTIMATE in stats.columns:
        stats[MED_POWER_ESTIMATE_W] = stats[POWER_ESTIMATE].rolling(MED_WINDOW, min_periods=MIN_PERIODS).median().clip(lower=0)
    stats.rename(columns={ELEVATION: ELEVATION_M}, inplace=True)

    if with_timespan:
        timespans = stats[TIMESPAN_ID].unique()
    stats['keep'] = pd.notna(stats[HR_IMPULSE_10])
    stats.interpolate(method='time', inplace=True)
    stats = stats.loc[stats['keep'] == True].drop(columns=['keep'])
    if with_timespan:
        stats = stats.loc[stats[TIMESPAN_ID].isin(timespans)]

    stats[CLIMB_MS] = stats[ELEVATION_M].diff() * 0.1
    stats[TIME] = pd.to_datetime(stats.index)
    stats[LOCAL_TIME] = stats[TIME].apply(lambda x: time_to_local_time(x.to_pydatetime(), HMS))

    return stats


def std_health_statistics(s, *extra, start=None, finish=None):

    from ..stoats.calculate.monitor import MonitorCalculator

    # this assumes FF cover all the dates and HR/steps fit into them.  may not be true in all cases?
    # also, we downsample the FF data to hourly intervals then shift daily data to match one of those times
    # this avoids introducing gaps in the FF data when merging that mess up the continuity of the plots.
    stats_1 = statistics(s, FITNESS_D_ANY, FATIGUE_D_ANY, start=start, finish=finish).resample('1h').mean()
    stats_2 = statistics(s, REST_HR, start=start, finish=finish, owner=MonitorCalculator). \
        reindex(stats_1.index, method='nearest', tolerance=dt.timedelta(minutes=30))
    stats_3 = statistics(s, DAILY_STEPS, ACTIVE_TIME, ACTIVE_DISTANCE, *extra, start=start, finish=finish). \
        reindex(stats_1.index, method='nearest', tolerance=dt.timedelta(minutes=30))
    stats = stats_1.merge(stats_2, how='outer', left_index=True, right_index=True)
    stats = stats.merge(stats_3, how='outer', left_index=True, right_index=True)
    for fitness in like(FITNESS_D_ANY, stats.columns):
        stats[_log(fitness)] = np.log10(stats[fitness])
    for fatigue in like(FATIGUE_D_ANY, stats.columns):
        stats[_log(fatigue)] = np.log10(stats[fatigue])
    stats[ACTIVE_TIME_H] = stats[ACTIVE_TIME] / 3600
    stats[ACTIVE_DISTANCE_KM] = stats[ACTIVE_DISTANCE] / 1000
    stats[TIME] = pd.to_datetime(stats.index)
    stats[LOCAL_TIME] = stats[TIME].apply(lambda x: time_to_local_time(x.to_pydatetime(), YMD))

    return stats


def nearby_activities(s, local_time=None, time=None, activity_journal_id=None, activity_group_name=None):
    activity_journal_id = activity_journal(s, local_time, time, activity_journal_id,
                                           activity_group_name=activity_group_name)
    return nearby_any_time(s, ActivityJournal.from_id(s, activity_journal_id))


def bookmarks(s, constraint, owner=CoastingBookmark):
    yield from s.query(ActivityBookmark). \
        filter(ActivityBookmark.owner == owner,
               ActivityBookmark.constraint == constraint).all()


def statistic_names(s, *statistics, owner=None, constraint=None, check=True):
    unresolved = [statistic for statistic in statistics if not isinstance(statistic, StatisticName)]
    if unresolved:
        q = s.query(StatisticName). \
            filter(or_(StatisticName.name.like(statistic) for statistic in unresolved))
        if owner:
            q = q.filter(StatisticName.owner == owner)
        if constraint:
            q = q.filter(StatisticName.constraint == constraint)
        resolved = q.all()
    else:
        resolved = []
    some = [statistic for statistic in statistics if isinstance(statistic, StatisticName)] + resolved
    if check and not some:
        raise Exception(f'Found no statistics for {statistics} (owner {owner}; constraint {constraint})')
    return some

_statistic_names = statistic_names


def _type_to_journal(t):
    return {StatisticJournalType.INTEGER: t.sji,
            StatisticJournalType.FLOAT: t.sjf,
            StatisticJournalType.TEXT: t.sjt}


def statistics(s, *statistics, start=None, finish=None, owner=None, constraint=None, sources=None,
               with_sources=False, check=True):
    t = _tables()
    ttj = _type_to_journal(t)
    names = statistic_names(s, *statistics, owner=owner, constraint=constraint, check=check)
    counts = Counter(name.name for name in names)
    labels = [name.name if counts[name.name] == 1 else f'{name.name} ({name.constraint})' for name in names]
    tables = [ttj[name.statistic_journal_type] for name in names]
    time_select = select([distinct(t.sj.c.time).label("time")]).select_from(t.sj). \
        where(t.sj.c.statistic_name_id.in_([n.id for n in names]))
    if start:
        time_select = time_select.where(t.sj.c.time >= start)
    if finish:
        time_select = time_select.where(t.sj.c.time <= finish)
    time_select = time_select.order_by("time").alias("sub_time")  # order here avoids extra index

    def sub_select(name, table):
        selects = [table.c.value, t.sj.c.time]
        if with_sources:
            selects += [t.sj.c.source_id]
        q = select(selects). \
                   select_from(t.sj.join(table)). \
                   where(t.sj.c.statistic_name_id == name.id)
        if sources:
            q = q.where(t.sj.c.source_id.in_(source.id for source in sources))
        # order_by doesn't affect plan but seems to speed up query
        return q.order_by(t.sj.c.time).alias(f'sub_{name.name}_{name.constraint}')

    sub_selects = [sub_select(name, table) for name, table in zip(names, tables)]
    # don't call this TIME because even though it's moved to index it somehow blocks the later addition
    # of a TIME column (eg when plotting health statistics)
    selects = [time_select.c.time.label(INDEX)] + \
              [sub.c.value.label(label) for sub, label in zip(sub_selects, labels)]
    if with_sources:
        selects += [sub.c.source_id.label(_src(label)) for sub, label in zip(sub_selects, labels)]
    sources = time_select
    for sub in sub_selects:
        sources = sources.outerjoin(sub, time_select.c.time == sub.c.time)
    sql = select(selects).select_from(sources)
    return pd.read_sql_query(sql=sql, con=s.connection(), index_col=INDEX)


def present(df, *names):
    if hasattr(df, 'columns'):
        return df is not None and all(name in df.columns and len(df[name].dropna()) for name in names)
    else:
        return df is not None and (len(df.dropna()) and all(df.name == name for name in names))


def median_d(df):
    return pd.Series(df.index).diff().median()


KEEP = 'keep'


def linear_resample(df, start=None, finish=None, d=None, quantise=True):
    log.debug(f'Linear resample with type {type(df.index)}, columns {df.columns}')
    d = d or median_d(df)
    start = start or df.index.min()
    finish = finish or df.index.max()
    if quantise:
        start, finish = int(start / d) * d, (1 + int(finish / d)) * d
    lin = pd.DataFrame({KEEP: True}, index=np.arange(start, finish, d))
    ldf = df.join(lin, how='outer', sort=True)
    # if this fails check for time-like columns
    ldf.interpolate(method='slinear', limit_area='inside', inplace=True)
    ldf = ldf.loc[ldf[KEEP] == True].drop(columns=[KEEP])
    return ldf


def median_dt(df):
    return pd.Series(df.index).diff().median().total_seconds()


def linear_resample_time(df, start=None, finish=None, dt=None, with_timespan=None, keep_nan=True, add_time=True):
    log.debug(f'Linear resample with type {type(df.index)}, columns {df.columns}')
    if with_timespan is None: with_timespan = TIMESPAN_ID in df.columns
    dt = dt or median_dt(df)
    start = start or df.index.min()
    finish = finish or df.index.max()
    lin = pd.DataFrame({KEEP: True}, index=pd.date_range(start=start, end=finish, freq=f'{dt}S'))
    ldf = df.copy().join(lin, how='outer', sort=True)
    # if this fails check for time-like columns
    ldf.interpolate(method='index', limit_area='inside', inplace=True)
    ldf = ldf.loc[ldf[KEEP] == True].drop(columns=[KEEP])
    if add_time:
        ldf[TIME] = ldf.index
        ldf[DELTA_TIME] = ldf[TIME].diff()
    if with_timespan:
        if keep_nan:
            ldf.loc[~ldf[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())] = np.nan
        else:
            ldf = ldf.loc[ldf[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())]
    return ldf


def groups_by_time(s, start=None, finish=None):
    q = s.query(ActivityJournal.start.label(INDEX), ActivityNearby.group.label(GROUP)). \
        filter(ActivityNearby.activity_journal_id == ActivityJournal.id)
    if start:
        q = q.filter(ActivityJournal.start >= start)
    if finish:
        q = q.filter(ActivityJournal.start < finish)
    return pd.read_sql_query(sql=q.statement, con=s.connection(), index_col=INDEX)




from collections import namedtuple
from json import loads
from logging import getLogger

from sqlalchemy.sql.functions import count

from . import UniProcCalculator, DataFrameCalculatorMixin
from ..load import StatisticJournalLoader
from ...data.frame import statistics
from ...data.impulse import pre_calc, DecayModel, calc
from ...lib.date import round_hour, to_time, local_date_to_time
from ...squeal import StatisticJournal, Composite, StatisticName, Source, Constant, CompositeComponent, \
    StatisticJournalFloat
from ...squeal.utils import add
from ...stoats.calculate.heart_rate import HRImpulse
from ...stoats.names import _src

log = getLogger(__name__)

Response = namedtuple('Response', 'src_name, src_owner, dest_name, tau_days, scale, start')


class ImpulseCalculator(DataFrameCalculatorMixin, UniProcCalculator):
    '''
    this is hard to do correctly, incrementally.

    for now, we do all or nothing.

    first, we check if the current solution is complete:
    * extends across full date range
    * all sources used
    (no need to check for gaps - composite chaining should do that)

    if not complete, we regenerate the whole damn thing.
    '''

    def __init__(self, *args, responses=None, impulse=None, **kargs):
        self.responses_ref = self._assert('responses', responses)
        self.impulse_ref = self._assert('impulse', impulse)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self.constants = [Constant.get(s, response) for response in self.responses_ref]
        self.responses = [Response(**loads(constant.at(s).value)) for constant in self.constants]
        self.impulse = HRImpulse(**loads(Constant.get(s, self.impulse_ref).at(s).value))

    def _delete(self, s):
        start, finish = self._start_finish(local_date_to_time)
        if start or self.force:
            if finish:
                log.warning(f'Ignoring finish - will delete all impulses from {start} onwards')
            self._delete_from(s, start)

    def _delete_from(self, s, start=None, inclusive=True):
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        if start:
            if inclusive:
                composite_ids = composite_ids.filter(StatisticJournal.time >= start)
            else:
                composite_ids = composite_ids.filter(StatisticJournal.time > start)
        log.debug(f'Delete query: {composite_ids}')
        n = s.query(count(Source.id)). \
            filter(Source.id.in_(composite_ids)). \
            scalar()
        if n:
            log.warning(f'Deleting {n} Composite sources ({start} onwards{" inclusive" if inclusive else ""})')
            s.query(Source). \
                filter(Source.id.in_(composite_ids)). \
                delete(synchronize_session=False)
            s.commit()

    def _missing(self, s):
        # clean out any gaps by unzipping the chained composites
        Composite.clean(s)
        # range we expect data for
        start = round_hour(self.__full_range(s, True), up=False)
        finish = round_hour(self.__full_range(s, False), up=True)
        n_secs = (finish - start).total_seconds()
        missing_coverage = any(self.__missing_coverage(s, response.dest_name, n_secs) for response in self.responses)
        missing_sources = self.__missing_sources(s)
        if missing_coverage or missing_sources:
            if missing_coverage: log.info('Incomplete coverage (so will re-calculate)')
            if missing_sources: log.info('Additional sources (so will re-calculate)')
            self._delete_from(s)
            return [(start, finish)]
        else:
            return []

    def __missing_coverage(self, s, response, n_secs):
        q = s.query(StatisticJournal.time). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == response,
                   StatisticName.owner == self.owner_out)
        start = q.order_by(StatisticJournal.time.asc()).limit(1).one_or_none()
        finish = q.order_by(StatisticJournal.time.desc()).limit(1).one_or_none()
        if start is None or finish is None:
            log.debug(f'Missing data for {response}')
            return True
        delta = (finish[0] - start[0]).total_seconds()
        if delta < n_secs:
            log.debug(f'Insufficient data for {response}: {delta}s / {n_secs}s (difference of {n_secs - delta}s)')
            return True
        return False

    def __missing_sources(self, s):
        response_ids = s.query(StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        inputs = s.query(CompositeComponent.input_source_id). \
            join(StatisticJournal, StatisticJournal.source_id == CompositeComponent.output_source_id). \
            filter(StatisticJournal.statistic_name_id.in_(response_ids))
        unused = s.query(count(StatisticJournal.id)). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.name == self.impulse.dest_name,
                   StatisticName.owner == self.owner_in,
                   ~StatisticJournal.source_id.in_(inputs)).scalar()
        return unused

    def __full_range(self, s, start=True):
        q = s.query(StatisticJournal.time). \
            filter(StatisticJournal.time > 1.0)  # avoid constants
        if start:
            q = q.order_by(StatisticJournal.time.asc())
        else:
            q = q.order_by(StatisticJournal.time.desc())
        return q.limit(1).scalar()

    def _run_one(self, s, missed):
        start, finish = missed
        hr10 = statistics(s, self.impulse.dest_name, owner=self.owner_in, with_sources=True, check=False)
        if not hr10.empty:
            all_sources = list(self.__make_sources(s, hr10))
            for response in self.responses:
                log.info(f'Creating values for {response.dest_name}')
                model = DecayModel(start=response.start, zero=0, scale=response.scale,
                                   period=response.tau_days * 24 * 60 * 60 / 3600,  # convert to intervals
                                   input=self.impulse.dest_name, output=response.dest_name)
                hr3600 = pre_calc(hr10.copy(), model, start=start, finish=finish)
                result = calc(hr3600, model)
                loader = StatisticJournalLoader(s, self.owner_out, add_serial=False)
                source, sources = None, list(all_sources)
                for time, row in result.iterrows():
                    while sources and time >= sources[0][0]:
                        source = sources.pop(0)[1]
                    loader.add(response.dest_name, None, None, None, source, row[response.dest_name], time,
                               StatisticJournalFloat)
                loader.load()

    def __make_sources(self, s, hr10):
        log.info('Creating sources')
        name = _src(self.impulse.dest_name)
        prev = add(s, Composite(n_components=0))
        yield to_time(0.0), prev
        for time, row in hr10.loc[hr10[name].ne(hr10[name].shift())].iterrows():
            id = row[name]
            composite = add(s, Composite(n_components=2))
            add(s, CompositeComponent(input_source_id=id, output_source=composite))
            add(s, CompositeComponent(input_source=prev, output_source=composite))
            yield time, composite
            prev = composite
        s.commit()




from math import sqrt
from random import uniform, gauss, seed, randrange
from time import time
from unittest import TestCase

from ch2.arty.spherical import Global
from ch2.arty.tree import CLRTree, MatchType, CQRTree, CERTree, LQRTree


class TestArty(TestCase):

    def known_points(self, tree):
        tree.add([(0, 0)], '0')
        tree.add([(0, 1)], 'y')
        tree.add([(1, 0)], 'x')
        tree.add([(1, 1)], '1')
        self.assertEqual(len(tree), 4)
        all = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINS))
        self.assertEqual(len(all), 4)
        some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINS))
        self.assertEqual(len(some), 2)
        some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINED))
        self.assertEqual(len(some), 0)
        some = list(tree.get([(0, 0), (0, 1)], match=MatchType.EQUALS))
        self.assertEqual(len(some), 0)
        self.assertEqual(len(tree), 4)
        tree.delete([(0, 0)])
        self.assertEqual(len(tree), 3)
        some = list(tree.get([(0, 0), (0, 1)], match=MatchType.CONTAINS))
        self.assertEqual(len(some), 1)

    def test_known_points(self):
        self.known_points(CLRTree(max_entries=2))
        self.known_points(CLRTree())
        self.known_points(CQRTree(max_entries=2))
        self.known_points(CQRTree())
        self.known_points(CERTree(max_entries=2))
        self.known_points(CERTree())

    def known_boxes(self, tree):
        tree.add([(-0.1, -0.1), (0.1, 0.1)], '0')
        tree.add([(0.9, 0.9), (1.1, 1.1)], '1')
        tree.add([(0, 0), (1, 1)], 'sq')
        tree.add([(2, 2), (3, 3)], 'x')
        some = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINED))
        self.assertEqual(len(some), 1)
        some = list(tree.get([(0, 0), (1, 1)], match=MatchType.CONTAINS))
        self.assertEqual(len(some), 1)
        some = list(tree.get([(0, 0), (2, 1)], match=MatchType.CONTAINED))
        self.assertEqual(len(some), 0)
        some = list(tree.get([(0, 0), (2, 1)], match=MatchType.CONTAINS))
        self.assertEqual(len(some), 1)
        some = list(tree.get([(0, 0), (1, 1)], match=MatchType.OVERLAP))
        self.assertEqual(len(some), 3)

    def test_known_boxes(self):
        self.known_boxes(CLRTree(max_entries=2))
        self.known_boxes(CLRTree())
        self.known_boxes(CQRTree(max_entries=2))
        self.known_boxes(CQRTree())
        self.known_points(CERTree(max_entries=2))
        self.known_points(CERTree())

    def test_delete_points(self):

        for size in 2, 3, 4, 8:

            def new_tree():
                tree = CQRTree(max_entries=size)
                for i in range(4):
                    for j in range(4):
                        tree.add([(i, j)], i+j)
                assert len(tree) == 16
                return tree

            tree = new_tree()
            tree.delete([(1, 2)])
            self.assertEqual(len(tree), 15)

            tree = new_tree()
            tree.delete([(1, 1), (2, 2)])
            self.assertEqual(len(tree), 16)  # default is EQUALS
            tree = new_tree()
            tree.delete([(1, 1), (2, 2)], match=MatchType.CONTAINS)
            self.assertEqual(len(tree), 12)
            tree = new_tree()
            tree.delete([(1, 1), (2, 2)], match=MatchType.CONTAINED)
            self.assertEqual(len(tree), 16)

    def random_box(self, n, size):
        x = uniform(0, size)
        y = uniform(0, size)
        dx = gauss(0, size/sqrt(n))
        dy = gauss(0, size/sqrt(n))
        return [(x, y), (x + dx, y + dy)]

    def gen_random(self, n, size=100):
        for i in range(n):
            yield 1, self.random_box(n, size=size)

    def test_equals(self, ):
        for size in 2, 3, 4, 8:
            tree1 = CQRTree(max_entries=size)
            for i, box in self.gen_random(10):
                tree1.add(box, i)
            tree2 = CQRTree(max_entries=size)
            for k, v in tree1.items():
                tree2.add(k, v)
            self.assertEqual(tree1, tree2)

    def test_underscores(self):

        for size in 2, 3, 4, 8:
            seed(size)
            tree1 = CQRTree(max_entries=size)
            for i, box in self.gen_random(10):
                tree1[box] = i
            tree2 = CQRTree(max_entries=size)
            for k, v in tree1.items():
                tree2.add(k, v)
            self.assertEqual(tree1, tree2)
            if size == 2:
                self.assertEqual(str(tree1), 'Quadratic RTree (10 leaves, 4 height, 1-2 entries)')
            tree3 = CQRTree(tree1.items(), max_entries=size)
            self.assertEqual(tree1, tree3)

        tree = CQRTree()
        tree.add([(1, 1)], '1')
        tree.add([(2, 2)], '2')
        self.assertEqual(len(tree), 2)
        self.assertTrue([(1, 1)] in tree)
        self.assertFalse([(3, 3)] in tree)
        self.assertTrue('1' in list(tree.values()))
        self.assertTrue(((1, 1),) in list(tree.keys()), list(tree.keys()))
        self.assertTrue((((1, 1),), '1') in list(tree.items()))
        del tree[[(1, 1)]]
        self.assertEqual(len(tree), 1)
        self.assertTrue(list(tree[[(2, 2)]]))
        self.assertTrue(list(tree.get([(2, 2)])))
        self.assertTrue(list(tree.get([(2, 2)], value='2')))
        self.assertFalse(list(tree.get([(2, 2)], value='1')))
        self.assertTrue(tree)
        del tree[[(2, 2)]]
        self.assertFalse(tree)

    def best_bug(self, tree):
        for i, (value, box) in enumerate(self.gen_random(100)):
            tree.add(box, value % 10)
            tree.assert_consistent()

    def test_best_bug(self):
        seed(4)  # 1:46 2:17 3:56 4:8 5:10 6:16 7:58 8:8
        self.best_bug(CLRTree(max_entries=2))
        self.best_bug(CQRTree(max_entries=2))
        self.best_bug(CERTree(max_entries=2))

    def stress(self, type, n_children, n_data, check=True):
        seed(1)
        tree = type(max_entries=n_children)
        data = list(self.gen_random(n_data))

        for value, box in data:
            tree.add(box, value)
            if check:
                tree.assert_consistent()

        for i in range(10):

            n_delete = randrange(int(1.1 * n_data))   # delete to empty 10% of time
            for j in range(n_delete):
                if data:
                    index = randrange(len(data))
                    value, box = data[index]
                    del data[index]
                    tree.delete_one(box, value=value)
                    if check:
                        tree.assert_consistent()
                else:
                    assert not tree, tree

            while len(data) < n_data:
                value, box = next(self.gen_random(1))
                data.append((value, box))
                tree.add(box, value)
                if check:
                    tree.assert_consistent()

            for j in range(n_data // 4):
                for match in range(4):
                    list(tree.get(self.random_box(10, 100), match=MatchType(match)))

    def test_stress(self):
        for type in CLRTree, CQRTree, CERTree:
            print('type %s' % type)
            for n_children in 3, 4, 10:
                print('n_children %d' % n_children)
                for n_data in 1, 2, 3, 100:
                    print('n_data %d' % n_data)
                    self.stress(type, n_children, n_data)

    def test_latlon(self):
        tree = LQRTree()
        for lon in -180, 180:
            tree.add([(lon, 0)], str(lon))
        for lon in -180, 180:
            found = list(tree.get([(lon, 0)]))
            self.assertEqual(len(found), 2)
        area = tree._area_of_mbr(tree._mbr_of_points(tree._normalize_points([(-179, -1), (179, 1)])))
        self.assertEqual(area, 4)

        tree = LQRTree()
        tree.add([(180, 0)], None)
        self.assertTrue(((180, 0),) in list(tree.keys()))

    def test_docs(self):
        tree = CQRTree()
        square = ((0,0),(0,1),(1,1),(1,0))
        tree[square] = 'square'
        self.assertEqual(list(tree[square]), ['square'])
        self.assertTrue(square in tree)
        diagonal = ((0,0),(1,1))
        self.assertEqual(list(tree[diagonal]), [])
        self.assertFalse(diagonal in tree)
        self.assertEqual(list(tree.keys()), [((0,0),(0,1),(1,1),(1,0))])
        self.assertEqual(list(tree.values()), ['square'])
        self.assertEqual(list(tree.items()), [(((0,0),(0,1),(1,1),(1,0)), 'square')])
        self.assertEqual(len(tree), 1)
        del tree[square]
        self.assertEqual(len(tree), 0)

        tree = CQRTree(default_match=MatchType.OVERLAP)
        tree[square] = 'square'
        self.assertEqual(list(tree[diagonal]), ['square'])

        tree = CQRTree(default_match=MatchType.OVERLAP)
        tree[square] = 'square'
        self.assertEqual(list(tree.get_items(diagonal)), [(((0,0),(0,1),(1,1),(1,0)), 'square')])

    def test_canary(self):
        tree = CQRTree()
        for i in range(5):
            tree.add([(i, i)], i)
        keys = tree.keys()
        next(keys)
        del tree[[(1, 1)]]
        with self.assertRaisesRegex(RuntimeError, 'mutated'):
            next(keys)

    def measure(self, tree, n_data, n_loops, dim=100):
        seed(1)
        data = list(self.gen_random(n_data, size=dim))
        start = time()
        for i in range(n_loops):
            for value, box in data:
                tree[box] = value
            assert len(tree) == len(data), len(tree)
            for _, box in data:
                list(tree[box])
            for value, box in data:
                del tree[box]
            assert len(tree) == 0
        return time() - start

    def measure_sizes(self):
        for type in CLRTree, CQRTree, CERTree:
            print()
            for size in 2, 3, 4, 6, 8, 10, 16, 32, 64, 128:
                for subtrees in True, False:
                    t = self.measure(type(max_entries=size, subtrees_flag=subtrees), 1000, 2)
                    print('%s %d %s %s' % (type, size, subtrees, t))
                    if t > 5 and size > 4:
                        print('abort')
                        return

    def test_global(self):

        def test_point(x, y, z):
            t = Global()
            t.add([(x, y)], z)
            l = list(t.get_items([(x, y)]))
            self.assertEqual(len(l), 9)
            for p, q in l:
                self.assertTrue(-0.001 < p[0][0] - x < 0.001)
                self.assertTrue(-0.001 < p[0][1] - y < 0.001)
                self.assertEqual(q, z)

        test_point(0.01, 0.01, 0)
        test_point(179.9, 0.01, 1)
        test_point(-179.9, 0.01, 2)
        test_point(0.01, 89.99, 0)
        test_point(179.9, 89.99, 1)
        test_point(-179.9, 89.99, 2)
        test_point(0.01, -89.99, 0)
        test_point(179.9, -89.99, 1)
        test_point(-179.9, -89.99, 2)

    def run_python(self, tree):
        tree[[(0, 0)]] = 'alice'
        tree[[(10, 10)]] = 'bob'
        tree[[(10, 10)]] = 'charles'

        def show(tree, x, y):
            found = False
            print()
            for entry in tree[[(x, y)]]:
                found = True
                print('%s at (%g,%g)' % (entry, x, y))
            if not found:
                print('nobody at (%g,%g)' % (x, y))

        show(tree, 0, 0)
        show(tree, 5, 5)
        show(tree, 10, 10)

        return tree

    def test_python(self):

        # for other-projects.md
        from ch2.arty import CQRTree, MatchType

        tree = self.run_python(CQRTree())
        self.assertEqual(list(tree[[(0, 0)]]), ['alice'])

        self.run_python(CQRTree(default_match=MatchType.OVERLAP, default_border=3))





from collections import defaultdict, namedtuple
from itertools import groupby
from json import loads
from logging import getLogger
from random import uniform

from sqlalchemy import inspect, select, alias, and_, distinct, func, not_, or_
from sqlalchemy.sql.functions import count

from . import UniProcCalculator
from ..names import LONGITUDE, LATITUDE, ACTIVE_DISTANCE
from ...arty import MatchType
from ...arty.spherical import SQRTree
from ...lib.date import to_time, local_date_to_time
from ...lib.dbscan import DBSCAN
from ...lib.optimizn import expand_max
from ...squeal import ActivityJournal, ActivityGroup, Constant, ActivitySimilarity, ActivityNearby, StatisticName, \
    StatisticJournal, StatisticJournalFloat, Timestamp

log = getLogger(__name__)
Nearby = namedtuple('Nearby', 'constraint, activity_group, border, start, finish, '
                              'latitude, longitude, height, width, fraction')


class SimilarityCalculator(UniProcCalculator):

    def __init__(self, *args, nearby=None, **kargs):
        self.nearby_ref = self._assert('nearby', nearby)
        super().__init__(*args, **kargs)

    def _startup(self, s):
        self.nearby = Nearby(**loads(Constant.get(s, self.nearby_ref).at(s).value))
        log.info(f'{self.nearby_ref}: {self.nearby}')
        log.info(f'Reducing to {int(0.5 + 100 * self.nearby.fraction):d}%')

    def _missing(self, s):
        prev = Timestamp.get(s, self.owner_out, self.nearby.constraint)
        if not prev:
            return [True]
        prev_ids = s.query(Timestamp.key). \
            filter(Timestamp.owner == self.owner_in,
                   Timestamp.constraint == None,
                   Timestamp.time < prev.time).cte()
        later = s.query(count(ActivityJournal.id)). \
            join(ActivityGroup). \
            filter(ActivityGroup.name == self.nearby.activity_group,
                   not_(ActivityJournal.id.in_(prev_ids))).scalar()
        if later:
            return [True]
        else:
            return []

    def _delete(self, s):
        start, finish = self._start_finish()
        log.warning(f'Deleting similarity data for {self.nearby.constraint} from {start} to {finish}')
        activity_ids = s.query(ActivityJournal.id)
        if start:
            activity_ids = activity_ids.filter(ActivityJournal.start >= local_date_to_time(start))
        if finish:
            activity_ids = activity_ids.filter(ActivityJournal.start < local_date_to_time(finish))
        s.query(ActivitySimilarity). \
            filter(ActivitySimilarity.constraint == self.nearby.constraint,
                   or_(ActivitySimilarity.activity_journal_lo_id.in_(activity_ids.cte()),
                       ActivitySimilarity.activity_journal_hi_id.in_(activity_ids.cte()))). \
            delete(synchronize_session=False)
        Timestamp.clear(s, self.owner_out, self.nearby.constraint)
        s.commit()

    def _run_one(self, s, missed):
        rtree = SQRTree(default_match=MatchType.OVERLAP, default_border=self.nearby.border)
        n_points = defaultdict(lambda: 0)
        self._prepare(s, rtree, n_points, 30000)
        n_overlaps = defaultdict(lambda: defaultdict(lambda: 0))
        new_ids, affected_ids = self._count_overlaps(s, rtree, n_points, n_overlaps, 10000)
        # this clears itself beforehand
        # use explicit class to distinguish from subclasses (which compare against this)
        with Timestamp(owner=self.owner_out, constraint=self.nearby.constraint).on_success(log, s):
            self._save(s, new_ids, affected_ids, n_points, n_overlaps, 10000)

    def _prepare(self, s, rtree, n_points, delta):
        n = 0
        for aj_id_in, lon, lat in self._filter(self._aj_lon_lat(s, new=False)):
            posn = [(lon, lat)]
            rtree[posn] = aj_id_in
            n_points[aj_id_in] += 1
            n += 1
            if n % delta == 0:
                log.info('Loaded %s points for %s' % (n, self.nearby.constraint))
        if n % delta:
            log.info('Loaded %s points for %s' % (n, self.nearby.constraint))

    def _count_overlaps(self, s, rtree, n_points, n_overlaps, delta):
        new_aj_ids, affected_aj_ids, n, no = [], set(), 0, 0
        for aj_id_in, aj_lon_lats in groupby(self._aj_lon_lat(s, new=True), key=lambda aj_lon_lat: aj_lon_lat[0]):
            aj_lon_lats = list(self._filter(aj_lon_lats))  # reuse below
            seen_posns = set()
            new_aj_ids.append(aj_id_in)
            affected_aj_ids.add(aj_id_in)
            for _, lon, lat in aj_lon_lats:
                posn = [(lon, lat)]
                for other_posn, aj_id_out in rtree.get_items(posn):
                    if other_posn not in seen_posns:
                        lo, hi = min(aj_id_in, aj_id_out), max(aj_id_in, aj_id_out)  # ordered pair
                        affected_aj_ids.add(aj_id_out)
                        n_overlaps[lo][hi] += 1
                        no += 1
                        seen_posns.add(other_posn)
            for _, lon, lat in aj_lon_lats:  # adding after avoids matching ourselves
                posn = [(lon, lat)]
                rtree[posn] = aj_id_in
                n_points[aj_id_in] += 1
                n += 1
                if n % delta == 0:
                    log.info(f'Measured {n} points for {self.nearby.constraint} ({no} overlaps)')
        if n % delta:
            log.info('Measured %s points for %s' % (n, self.nearby.constraint))
        return new_aj_ids, affected_aj_ids

    def _filter(self, lon_lats):
        for lon_lat in lon_lats:
            if uniform(0, 1 / self.nearby.fraction) < 1:
                yield lon_lat

    def _aj_lon_lat(self, s, new=True):

        start = to_time(self.nearby.start)
        finish = to_time(self.nearby.finish)

        agroup = s.query(ActivityGroup.id).filter(ActivityGroup.name == self.nearby.activity_group).scalar()
        lat = s.query(StatisticName.id). \
            filter(StatisticName.name == LATITUDE, StatisticName.constraint == agroup).scalar()
        lon = s.query(StatisticName.id). \
            filter(StatisticName.name == LONGITUDE, StatisticName.constraint == agroup).scalar()

        sj_lat = inspect(StatisticJournal).local_table
        sj_lon = alias(inspect(StatisticJournal).local_table)
        sjf_lat = inspect(StatisticJournalFloat).local_table
        sjf_lon = alias(inspect(StatisticJournalFloat).local_table)
        aj = inspect(ActivityJournal).local_table
        ns = inspect(ActivitySimilarity).local_table

        existing_lo = select([ns.c.activity_journal_lo_id]). \
            where(ns.c.constraint == self.nearby.constraint)
        existing_hi = select([ns.c.activity_journal_hi_id]). \
            where(ns.c.constraint == self.nearby.constraint)
        existing = existing_lo.union(existing_hi).cte()

        stmt = select([sj_lat.c.source_id, sjf_lon.c.value, sjf_lat.c.value]). \
            select_from(sj_lat).select_from(sj_lon).select_from(sjf_lat).select_from(sjf_lat).select_from(aj). \
            where(and_(sj_lat.c.source_id == sj_lon.c.source_id,  # same source
                       sj_lat.c.time == sj_lon.c.time,            # same time
                       sj_lat.c.source_id == aj.c.id,             # and associated with an activity
                       aj.c.activity_group_id == agroup,          # of the right group
                       sj_lat.c.id == sjf_lat.c.id,               # lat sub-class
                       sj_lon.c.id == sjf_lon.c.id,               # lon sub-class
                       sj_lat.c.statistic_name_id == lat,         # lat name
                       sj_lon.c.statistic_name_id == lon,         # lon name
                       sj_lat.c.time >= start.timestamp(),        # time limits
                       sj_lat.c.time < finish.timestamp(),
                       sjf_lat.c.value > self.nearby.latitude - self.nearby.height / 2,
                       sjf_lat.c.value < self.nearby.latitude + self.nearby.height / 2,
                       sjf_lon.c.value > self.nearby.longitude - self.nearby.width / 2,
                       sjf_lon.c.value < self.nearby.longitude + self.nearby.width / 2))

        if new:
            stmt = stmt.where(func.not_(sj_lat.c.source_id.in_(existing)))
        else:
            stmt = stmt.where(sj_lat.c.source_id.in_(existing))
        stmt = stmt.order_by(sj_lat.c.source_id)  # needed for seen logic
        yield from s.connection().execute(stmt)

    def _save(self, s, new_ids, affected_ids, n_points, n_overlaps, delta):
        distances = dict((s.source.id, s.value)
                         for s in s.query(StatisticJournalFloat).
                         join(StatisticName).
                         filter(StatisticName.name == ACTIVE_DISTANCE,
                                StatisticName.owner == self.owner_in).all())  # todo - another owner
        n = 0
        for lo in affected_ids:
            add_lo, d_lo = lo in new_ids, distances.get(lo, None)
            if d_lo:
                for hi in (id for id in affected_ids if id > lo):
                    d_hi = distances.get(hi, None)
                    if d_hi and (add_lo or hi in new_ids):
                        if lo in new_ids and hi not in new_ids:
                            # hi already existed so was added first
                            n_max = n_points[hi]
                            d_factor = d_hi / d_lo if d_hi < d_lo else 1
                        else:
                            # both new and ordered lo-hi, or hi new and last
                            n_max = n_points[lo]
                            d_factor = d_lo / d_hi if d_lo < d_hi else 1
                        s.add(ActivitySimilarity(constraint=self.nearby.constraint,
                                                 activity_journal_lo_id=lo, activity_journal_hi_id=hi,
                                                 similarity=(n_overlaps[lo][hi] / n_max) * d_factor))
                        n += 1
                        if n % delta == 0:
                            log.info('Saved %d for %s' % (n, self.nearby.constraint))
        if n % delta:
            log.info('Wrote %d for %s' % (n, self.nearby.constraint))


class NearbySimilarityDBSCAN(DBSCAN):

    def __init__(self, s, constraint, epsilon, minpts):
        super().__init__(epsilon, minpts)
        self.__s = s
        self.__constraint = constraint
        self.__max_similarity = self.__s.query(func.max(ActivitySimilarity.similarity)). \
            filter(ActivitySimilarity.constraint == constraint).scalar()
        # self._log.info('Max similarity %.2f' % self.__max_similarity)

    def run(self):
        candidates = set(x[0] for x in
                         self.__s.query(distinct(ActivitySimilarity.activity_journal_lo_id)).
                         filter(ActivitySimilarity.constraint == self.__constraint).all())
        candidates.union(set(x[0] for x in
                             self.__s.query(distinct(ActivitySimilarity.activity_journal_lo_id)).
                             filter(ActivitySimilarity.constraint == self.__constraint).all()))
        candidates = sorted(candidates)
        # shuffle(candidates)  # skip for repeatability
        return super().run(candidates)

    def neighbourhood(self, candidate, epsilon):
        qlo = self.__s.query(ActivitySimilarity.activity_journal_lo_id). \
            filter(ActivitySimilarity.constraint == self.__constraint,
                   ActivitySimilarity.activity_journal_hi_id == candidate,
                   (self.__max_similarity - ActivitySimilarity.similarity) / self.__max_similarity < epsilon)
        qhi = self.__s.query(ActivitySimilarity.activity_journal_hi_id). \
            filter(ActivitySimilarity.constraint == self.__constraint,
                   ActivitySimilarity.activity_journal_lo_id == candidate,
                   (self.__max_similarity - ActivitySimilarity.similarity) / self.__max_similarity < epsilon)
        return [x[0] for x in qlo.all()] + [x[0] for x in qhi.all()]


class NearbyCalculator(UniProcCalculator):

    def __init__(self, *args, constraint=None, **kargs):
        self.constraint = self._assert('constraint', constraint)
        super().__init__(*args, **kargs)

    def _missing(self, s):
        latest_similarity = Timestamp.get(s, self.owner_in, self.constraint)
        latest_groups = Timestamp.get(s, self.owner_out, self.constraint)
        if not latest_groups or latest_similarity.time > latest_groups.time:
            self._delete(s)  # missing isn't really missing...
            return [True]
        else:
            return []

    def _delete(self, s):
        Timestamp.clear(s, self.owner_out, constraint=self.constraint)
        s.query(ActivityNearby).filter(ActivityNearby.constraint == self.constraint).delete()

    def _run_one(self, s, missed):
        with Timestamp(owner=self.owner_out, constraint=self.constraint).on_success(log, s):
            d_min, n = expand_max(log, 0, 1, 5, lambda d: len(self.dbscan(s, d)))
            log.info(f'{n} groups at d={d_min}')
            self.save(s, self.dbscan(s, d_min))

    def dbscan(self, s, d):
        return NearbySimilarityDBSCAN(s, self.constraint, d, 3).run()

    def save(self, s, groups):
        for i, group in enumerate(groups):
            log.info(f'Group {i} has {len(group)} members')
            for activity_journal_id in group:
                s.add(ActivityNearby(constraint=self.constraint, group=i,
                                     activity_journal_id=activity_journal_id))



import re

POW_M1 = '\u207b\u00b9'
POW_2 = '\u00b2'

ACTIVE_DISTANCE = 'Active Distance'
ACTIVE_SPEED = 'Active Speed'
ACTIVE_TIME = 'Active Time'
ACTIVITY = 'Activity'
AIR_SPEED = 'Air Speed'
ALTITUDE = 'Altitude'
ASPECT_RATIO = 'Aspect Ratio'
CADENCE = 'Cadence'
CALORIE_ESTIMATE = 'Calorie Estimate'
CDA = 'CdA'
CLIMB = 'Climb'
CLIMB_DISTANCE = 'Climb Distance'
CLIMB_ELEVATION = 'Climb Elevation'
CLIMB_GRADIENT = 'Climb Gradient'
CLIMB_TIME = 'Climb Time'
COLOR = 'Color'
CRR = 'Crr'
CUMULATIVE_STEPS = 'Cumulative Steps'
CUMULATIVE_STEPS_START = 'Cumulative Steps Start'
CUMULATIVE_STEPS_FINISH = 'Cumulative Steps Finish'
DAILY_STEPS = 'Daily Steps'
DELAYED_POWER = 'Delayed Power'
DETRENDED_HEART_RATE = 'Detrended Heart Rate'
DIRECTION = 'Direction'
DISTANCE = 'Distance'
DUMMY = 'Dummy'
ELEVATION = 'Elevation'
ENERGY = 'Energy'
ENERGY_ESTIMATE = 'Energy Estimate'
FATIGUE = 'Fatigue'
FINISH = 'Finish'
FITNESS = 'Fitness'
FATIGUE_D = 'Fatigue %dd'
FATIGUE_D_ANY = 'Fatigue %'
FITNESS_D = 'Fitness %dd'
FITNESS_D_ANY = 'Fitness %'
FTHR = 'FTHR'
GROUP = 'Group'
HEADING = 'Heading'
HEART_RATE = 'Heart Rate'
HR_IMPULSE = 'HR Impulse'
HR_IMPULSE_10 = 'HR Impulse / 10s'
HR_ZONE = 'HR Zone'
INDEX = 'Index'
LATITUDE = 'Latitude'
LOCAL_TIME = 'Date'
LONGITUDE = 'Longitude'
LOSS = 'Loss'
MAX_MEAN_PE_M = 'Max Mean PE %dm'
MAX_MEAN_PE_M_ANY = 'Max Mean PE %'
MAX_MED_HR_M = 'Max Med HR %dm'
MAX_MED_HR_M_ANY = 'Max Med HR %'
MEAN_POWER_ESTIMATE = 'Mean Power Estimate'
MED_KM_TIME = 'Med %dkm Time'
MED_KM_TIME_ANY = 'Med % Time'
MIN_KM_TIME = 'Min %dkm Time'
MIN_KM_TIME_ANY = 'Min % Time'
PERCENT_IN_Z = 'Percent in Z%d'
PERCENT_IN_Z_ANY = 'Percent in Z%'
POWER_ESTIMATE = 'Power Estimate'
PREDICTED_HEART_RATE = 'Predicted Heart Rate'
SEGMENT_TIME = 'Segment Time'
SEGMENT_DISTANCE = 'Segment Distance'
SEGMENT_HEART_RATE = 'Segment Heart Rate'
SOURCE = 'Source'
SPEED = 'Speed'
SPHERICAL_MERCATOR_X = 'Spherical Mercator X'
SPHERICAL_MERCATOR_Y = 'Spherical Mercator Y'
STEPS = 'Steps'
RAW_ELEVATION = 'Raw Elevation'
REST_HR = 'Rest HR'
TIME = 'Time'
TIME_IN_Z = 'Time in Z%d'
TIME_IN_Z_ANY = 'Time in Z%'
TIMESPAN_ID = 'Timespan ID'
TOTAL_CLIMB = 'Total Climb'


MAX = '[max]'
MIN = '[min]'
CNT = '[cnt]'
AVG = '[avg]'
SUM = '[sum]'
MSR = '[msr]'   # measure (separate table, less efficient)
def summaries(*args): return ','.join(args)


BPM = 'bpm'
DEG = 'deg'
ENTRIES = 'entries'
FF = 'FF'
H = 'h'
KJ = 'kJ'
KCAL = 'kCal'
KG = 'kg'
KM = 'km'
KMH = f'kmh{POW_M1}'
M = 'm'
MS = f'ms{POW_M1}'
PC = '%'
RPM = 'rpm'
S = 's'
STEPS_UNITS = 'steps'
W = 'W'


def _d(name): return f'Δ {name}'
def _avg(name): return f'Avg {name}'
def _cor(name): return f'Cor {name}'
def _slash(name, units): return f'{name} / {units}'
def _log(name): return f'Log {name}'
def _sqr(name): return f'{name}{POW_2}'
def _new(name): return f'New {name}'
def _src(name): return f'Src {name}'

MED_WINDOW = '60s'
def _med(name): return f'Med{MED_WINDOW} {name}'

def like(pattern, names):
    return list(_like(pattern, names))

def _like(pattern, names):
    matcher = re.compile(re.sub('%', '.+', pattern))
    for name in names:
        if matcher.match(name):
            yield name


AIR_SPEED_2 = _sqr(AIR_SPEED)
AVG_AIR_SPEED_2 = _avg(AIR_SPEED_2)
DELTA_AIR_SPEED_2 = _d(AIR_SPEED_2)

SPEED_2 = _sqr(SPEED)
AVG_SPEED_2 = _avg(SPEED_2)
DELTA_SPEED_2 = _d(SPEED_2)

DELTA_TIME = _d(TIME)
DELTA_DISTANCE = _d(DISTANCE)
DELTA_ELEVATION = _d(ELEVATION)
DELTA_SPEED = _d(SPEED)
DELTA_ENERGY = _d(ENERGY)
COR_HEART_RATE = _cor(HEART_RATE)

ACTIVE_DISTANCE_KM = _slash(ACTIVE_DISTANCE, KM)
ACTIVE_TIME_H = _slash(ACTIVE_TIME, H)
CLIMB_MS = _slash(CLIMB, MS)
DISTANCE_KM = _slash(DISTANCE, KM)
ELEVATION_M = _slash(ELEVATION, M)
HEART_RATE_BPM = _slash(HEART_RATE, BPM)
SPEED_KMH = _slash(SPEED, KMH)
POWER_ESTIMATE_W = _slash(POWER_ESTIMATE, W)

LOG_FITNESS = _log(FITNESS)
LOG_FATIGUE = _log(FATIGUE)
MED_SPEED_KMH = _med(SPEED_KMH)
MED_HEART_RATE_BPM = _med(HEART_RATE_BPM)
MED_HR_IMPULSE_10 = _med(HR_IMPULSE_10)
MED_CADENCE = _med(CADENCE)
MED_POWER_ESTIMATE_W = _med(POWER_ESTIMATE_W)

POWER_HR = 'Power / HR'
POWER_HR_LAG = 'Power / HR Lag'
HR_DRIFT = 'HR Drift'
LOG_HR_DRIFT = _log('HR Drift')
IDLE_HR = 'Idle HR'
WIND_SPEED = 'Wind Speed'
WIND_HEADING = 'Wind Heading'




from unittest import TestCase

from ch2.lib.date import to_time, format_time, local_time_to_time, time_to_local_time


class TestDate(TestCase):

    def test_to_datetime(self):
        self.assertEqual(format_time(to_time('2810-09-21 13:24:01')), '2810-09-21 13:24:01')
        self.assertEqual(format_time(to_time('2810-09-21 13:24')), '2810-09-21 13:24:00')
        self.assertEqual(format_time(to_time('2810-09-21')), '2810-09-21 00:00:00')

    def test_tz(self):
        time = local_time_to_time('2019-02-16 12:19:00')
        self.assertEqual(time, to_time('2019-02-16 15:19:00'))
        self.assertEqual(time_to_local_time(time), '2019-02-16 12:19:00')




import pandas as pd
from tempfile import NamedTemporaryFile
from unittest import TestCase

from ch2 import activities, constants
from ch2.commands.args import bootstrap_file, mm, DEV, FAST, V, m
from ch2.config import default, getLogger
from ch2.data import activity_statistics, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, \
    ELEVATION, SPEED, CADENCE, HEART_RATE, TIME, TIMESPAN_ID

log = getLogger(__name__)


class TestPower(TestCase):

    def test_constant(self):
        with NamedTemporaryFile() as f:

            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

            args, db = bootstrap_file(f, m(V), '5', 'constants', '--set', 'FTHR.%', '154')
            constants(args, db)
            args, db = bootstrap_file(f, m(V), '5', 'constants', 'FTHR.%')
            constants(args, db)
            args, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                           'activities', mm(FAST), 'data/test/source/personal/2018-03-04-qdp.fit')
            activities(args, db)

            with db.session_context() as s:
                stats = activity_statistics(s, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                            DISTANCE, ELEVATION, SPEED, CADENCE, HEART_RATE,
                                            local_time='2018-03-04 07:16:33', activity_group_name='Bike',
                                            with_timespan=True)
                stats.describe()

                sepn = pd.Series(stats.index).diff().median()  # 7 secs
                start = stats.index.min()  # 2018-03-04 10:16:33+00:00
                finish = stats.index.max()  # 2018-03-04 16:34:51+00:00
                even = pd.DataFrame({'keep': True}, index=pd.date_range(start=start, end=finish, freq=sepn))
                both = stats.join(even, how='outer', sort=True)
                both.interpolate(method='index', limit_area='inside', inplace=True)
                both = both.loc[both['keep'] == True].drop(columns=['keep'])
                both = both.loc[both[TIMESPAN_ID].isin(stats[TIMESPAN_ID].unique())]
                both.describe()




from logging import getLogger

from ..commands.args import PATH, FORCE, FAST, WORKER, parse_pairs, KARG
from ..squeal.tables.pipeline import PipelineType
from ..stoats.pipeline import run_pipeline

log = getLogger(__name__)


def monitor(args, db):
    '''
## monitor

    > ch2 monitor PATH [PATH ...]

Read monitor data from FIT files.

Note: When using bash use `shopt -s globstar` to enable ** globbing.
    '''
    run_pipeline(db, PipelineType.MONITOR, paths=args[PATH], force=args[FORCE],
                 worker=args[WORKER] is not None, id=args[WORKER], **parse_pairs(args[KARG]))
    if not args[FAST]:
        run_pipeline(db, PipelineType.STATISTIC)




from bokeh.models import PanTool, ZoomInTool, ZoomOutTool, ResetTool, HoverTool

from ...stoats.names import DISTANCE_KM, LOCAL_TIME


def tooltip(name):
    return name, '@{%s}' % name


def make_tools(y):
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool(),
             HoverTool(tooltips=[tooltip(x) for x in (y, DISTANCE_KM, LOCAL_TIME)])]
    return tools




from logging import getLogger

from .args import DIARY, no, SUB_COMMAND, DEFAULT, CHECK, DATA, CONFIG, ACTIVITY_GROUPS
from ..config import default, ActivityGroup
from ..squeal.tables.statistic import StatisticName, StatisticJournal

log = getLogger(__name__)


def config(args, db):
    '''
## config

    > ch2 config default

Generate a simple initial configuration.

Please see the documentation at http://andrewcooke.github.io/choochoo - you have a lot more options!

    > ch2 config check --no-config --no-data

Check that the current database is empty.
    '''
    action = args[SUB_COMMAND]
    if action == DEFAULT:
        no_diary = args[no(DIARY)]
        check(db, not no_diary, not no_diary, True)
        default(db, no_diary=no_diary)
    elif action == CHECK:
        check(db, args[no(CONFIG)], args[no(DATA)], args[no(ACTIVITY_GROUPS)])


def check(db, config, data, activity_groups):
    with db.session_context() as s:
        if config:
            if s.query(StatisticName).count():
                raise Exception('The database already contains StatisticName entries')
        if data:
            if s.query(StatisticJournal).count():
                raise Exception('The database already contains StatisticJournal entries')
        if activity_groups:
            if s.query(ActivityGroup).count():
                raise Exception('The database already contains ActivityGroup entries')




import datetime as dt
from abc import abstractmethod
from logging import getLogger

from sqlalchemy import not_
from sqlalchemy.sql.functions import count

from ..pipeline import MultiProcPipeline, UniProcPipeline, LoaderMixin
from ..waypoint import WaypointReader
from ...commands.args import STATISTICS, WORKER, mm
from ...lib.date import local_time_to_time, time_to_local_time, format_date, to_date, local_date_to_time, \
    time_to_local_date
from ...lib.log import log_current_exception
from ...lib.schedule import Schedule
from ...squeal import ActivityJournal, Interval, Timestamp, StatisticJournal, StatisticName, Composite, \
    CompositeComponent, Source
from ...squeal.types import long_cls
from ...squeal.utils import add

log = getLogger(__name__)


class CalculatorMixin:

    def __init__(self, *args, owner_in=None, start=None, finish=None, **kargs):
        self.owner_in = self._assert('owner_in', owner_in)  # todo - not needed by some...
        self.start = start  # optional start local time (always present for workers)
        self.finish = finish  # optional finish local time (always present for workers)
        super().__init__(*args, **kargs)

    def _start_finish(self, type=None):
        start, finish = self.start, self.finish
        if type:
            if start: start = type(start)
            if finish: finish = type(finish)
        return start, finish

    def _base_command(self):
        return f'{{ch2}} -v0 -l {{log}} -f {self._db.path} {STATISTICS} {mm(WORKER)} {self.id}'


class MultiProcCalculator(CalculatorMixin, MultiProcPipeline):

    pass


class UniProcCalculator(CalculatorMixin, UniProcPipeline):

    def _base_command(self):
        raise Exception('UniProc does not support workers')

    def _args(self, missing, start, finish):
        raise Exception('UniProc does not support workers')


class ActivityJournalCalculatorMixin:

    def _delimit_query(self, q):
        start, finish = self._start_finish(type=local_time_to_time)
        log.debug(f'Delimit times: {start} - {finish}')
        if start:
            q = q.filter(ActivityJournal.start >= start)
        if finish:
            q = q.filter(ActivityJournal.start <= finish)
        return q

    def _missing(self, s):
        existing_ids = s.query(Timestamp.key).filter(Timestamp.owner == self.owner_out)
        q = s.query(ActivityJournal.start). \
            filter(not_(ActivityJournal.id.in_(existing_ids.cte()))). \
            order_by(ActivityJournal.start)
        return [row[0] for row in self._delimit_query(q)]

    def _args(self, missing, start, finish):
        s, f = time_to_local_time(missing[start]), time_to_local_time(missing[finish])
        log.info(f'Starting worker for {s} - {f}')
        return f'"{s}" "{f}"'

    def _delete(self, s):
        start, finish = self._start_finish(type=local_time_to_time)
        s.commit()   # so that we don't have any risk of having something in the session that can be deleted
        statistic_names = s.query(StatisticName.id).filter(StatisticName.owner == self.owner_out)
        activity_journals = self._delimit_query(s.query(ActivityJournal.id))
        statistic_journals = s.query(StatisticJournal.id). \
            filter(StatisticJournal.statistic_name_id.in_(statistic_names.cte()),
                   StatisticJournal.source_id.in_(activity_journals))
        for repeat in range(2):
            if repeat:
                s.query(StatisticJournal).filter(StatisticJournal.id.in_(statistic_journals.cte())). \
                    delete(synchronize_session=False)
                Timestamp.clean_keys(log, s, activity_journals.cte(), self.owner_out, constraint=None)
            else:
                n = s.query(count(StatisticJournal.id)). \
                    filter(StatisticJournal.id.in_(statistic_journals.cte())).scalar()
                if n:
                    log.warning(f'Deleting {n} statistics for {long_cls(self.owner_out)} from {start} to {finish}')
                else:
                    log.warning(f'No statistics to delete for {long_cls(self.owner_out)} from {start} to {finish}')
                    # log.debug(statistic_journals)
        s.commit()

    def _get_source(self, s, time):
        return s.query(ActivityJournal).filter(ActivityJournal.start == time).one()


class DataFrameCalculatorMixin(LoaderMixin):

    def _run_one(self, s, time_or_date):
        source = self._get_source(s, time_or_date)
        with Timestamp(owner=self.owner_out, key=source.id).on_success(log, s):
            try:
                # data may be structured (doesn't have to be simply a dataframe)
                data = self._read_dataframe(s, source)
                stats = self._calculate_stats(s, source, data)
                loader = self._get_loader(s)
                self._copy_results(s, source, loader, stats)
                loader.load()
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}')
                log_current_exception()

    @abstractmethod
    def _get_source(self, s, time_or_date):
        raise NotImplementedError()

    @abstractmethod
    def _read_dataframe(self, s, source):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_stats(self, s, source, data):
        raise NotImplementedError()

    @abstractmethod
    def _copy_results(self, s, source, loader, stats):
        raise NotImplementedError()


class DirectCalculatorMixin(LoaderMixin):

    def _run_one(self, s, time_or_date):
        source = self._get_source(s, time_or_date)
        with Timestamp(owner=self.owner_out, key=source.id).on_success(log, s):
            try:
                data = self._read_data(s, source)
                loader = self._get_loader(s)
                self._calculate_results(s, source, data, loader)
                loader.load()
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}')
                log_current_exception()

    @abstractmethod
    def _get_source(self, s, time_or_date):
        raise NotImplementedError()

    @abstractmethod
    def _read_data(self, s, source):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, source, data, loader):
        raise NotImplementedError()


class WaypointCalculatorMixin(DirectCalculatorMixin):

    # todo - can / should this be replaced by a data-frame approach?

    def _read_data(self, s, source):
        waypoints = list(WaypointReader().read(s, source, self._names(), self._assert('owner_in', self.owner_in)))
        if not waypoints:
            raise Exception('No waypoints')
        else:
            return waypoints

    @abstractmethod
    def _names(self):
        raise NotImplementedError()


class IntervalCalculatorMixin(LoaderMixin):

    def __init__(self, *args, schedule='m', load_once=False, **kargs):
        self.schedule = Schedule(self._assert('schedule', schedule))
        self.load_once = load_once
        self._prev_loader = None
        super().__init__(*args, **kargs)

    def _missing(self, s):
        start, finish = self._start_finish(type=to_date)
        return list(Interval.missing_dates(log, s, self.schedule, self.owner_out, start=start, finish=finish))

    def _args(self, missing, start, finish):
        s, f = format_date(missing[start][0]), format_date(missing[finish][1])
        log.info(f'Starting worker for {s} - {f}')
        return f'"{s}" "{f}"'

    def _delete(self, s):
        start, finish = self._start_finish()
        # we delete the intervals that the statistics depend on and they will cascade
        for repeat in range(2):
            if repeat:
                q = s.query(Interval)
            else:
                q = s.query(count(Interval.id))
            q = q.filter(Interval.owner == self.owner_out)
            if start:
                q = q.filter(Interval.finish > start)
            if finish:
                q = q.filter(Interval.start < finish)
            if repeat:
                for interval in q.all():
                    log.debug('Deleting %s' % interval)
                    s.delete(interval)
            else:
                n = q.scalar()
                if n:
                    log.warning('Deleting %d intervals' % n)
                else:
                    log.warning('No intervals to delete')
        s.commit()

    def _run_one(self, s, missing):
        if s.query(Interval). \
            filter(Interval.schedule == self.schedule,
                   Interval.owner == self.owner_out,
                   Interval.start == missing[0]).one_or_none():
            raise Exception('Interval already exists')
        interval = add(s, Interval(schedule=self.schedule, owner=self.owner_out,
                                   start=missing[0], finish=missing[1]))
        s.commit()
        try:
            data = self._read_data(s, interval)
            if self.load_once and self._prev_loader:
                loader = self._prev_loader
            else:
                loader = self._get_loader(s, add_serial=False, clear_timestamp=False)
            self._calculate_results(s, interval, data, loader)
            if not self.load_once:
                loader.load()
            self._prev_loader = loader
        except Exception as e:
            log.warning(f'No statistics for {missing} due to error ({e})')
            log_current_exception()
            if not self.load_once:
                self._prev_loader = None

    def _shutdown(self, s):
        if self.load_once and self._prev_loader:
            log.debug('Single load')
            self._prev_loader.load()

    @abstractmethod
    def _read_data(self, s, interval):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, interval, data, loader):
        raise NotImplementedError()


class CompositeCalculatorMixin(LoaderMixin):

    # this assumes daily, but could probably be modified to use a schedule

    def _missing(self, s):
        dates = set()
        for source in self._unused_sources(s):
            dates.update(self._dates_for_source(source))
        start, finish = self._start_finish(lambda x: to_date(x, none=True))
        missing = sorted(filter(lambda d: (start is None or start <= d) and (finish is None or d <= finish), dates))
        if missing:
            log.debug(f'Missing {start} - {finish}: {missing[0]} - {missing[-1]} / {len(missing)}')
        return missing

    def _args(self, missing, start, finish):
        s, f = format_date(missing[start]), format_date(missing[finish])
        log.info(f'Starting worker for {s} - {f}')
        return f'"{s}" "{f}"'

    def _dates_for_source(self, mjournal):
        start = time_to_local_date(mjournal.start)
        finish = time_to_local_date(mjournal.finish) + dt.timedelta(days=1)
        while start < finish:
            yield start
            start += dt.timedelta(days=1)

    def _unused_sources(self, s):
        Composite.clean(s)
        used_sources = s.query(CompositeComponent.input_source_id). \
            join(StatisticJournal, CompositeComponent.output_source_id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        return self._unused_sources_given_used(s, used_sources)

    @abstractmethod
    def _unused_sources_given_used(self, s, used_sources):
        raise NotImplementedError()

    def _unused_sources_given_used_and_class(self, s, used_sources, source_cls):
        sources = s.query(source_cls).filter(~source_cls.id.in_(used_sources))
        start, finish = self._start_finish(local_date_to_time)
        if start:
            sources = sources.filter(source_cls.finish >= start)
        if finish:
            sources = sources.filter(source_cls.start <= finish)
        log.debug(f'Unused query: {sources}')
        return sources.all()

    def _delete(self, s):
        start, finish = self._start_finish(local_date_to_time)
        self._delete_time_range(s, start, finish)

    def _delete_time_range(self, s, start, finish):
        # separate this out because impulse needs to delete forwards
        composite_ids = s.query(Composite.id). \
            join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out)
        if start:
            composite_ids = composite_ids.filter(StatisticJournal.time >= start)
        if finish:
            composite_ids = composite_ids.filter(StatisticJournal.time <= finish)
        log.debug(f'Delete query: {composite_ids}')
        n = s.query(count(Source.id)). \
            filter(Source.id.in_(composite_ids)). \
            scalar()
        if n:
            log.warning(f'Deleting {n} Composite sources ({start} - {finish})')
            s.query(Source). \
                filter(Source.id.in_(composite_ids)). \
                delete(synchronize_session=False)
            s.commit()

    def _run_one(self, s, start):
        # careful here to make summertime work correctly
        finish = local_date_to_time(start + dt.timedelta(days=1))
        start = local_date_to_time(start)
        # in some cases we can have composites that need re-generating with additional data
        if s.query(count(Composite.id)). \
                join(StatisticJournal, Composite.id == StatisticJournal.source_id). \
                join(StatisticName). \
                filter(StatisticJournal.time >= start,
                       StatisticJournal.time < finish,
                       StatisticName.owner == self.owner_out).scalar():
            self._delete_time_range(s, start, finish)
        try:
            input_source_ids, data = self._read_data(s, start, finish)
            if not input_source_ids:
                raise Exception(f'No sources found on {start}')
            output_source = add(s, Composite(n_components=len(input_source_ids)))
            for input_source_id in input_source_ids:
                s.add(CompositeComponent(input_source_id=input_source_id, output_source=output_source))
            s.commit()
            loader = self._get_loader(s, add_serial=False, clear_timestamp=False)
            self._calculate_results(s, output_source, data, loader, start, finish)
            loader.load()
            self._prev_loader = loader
        except Exception as e:
            log.warning(f'No statistics for {start} - {finish} due to error ({e})')
            log_current_exception()

    @abstractmethod
    def _read_data(self, s, start, finish):
        raise NotImplementedError()

    @abstractmethod
    def _calculate_results(self, s, source, data, loader, start, finish):
        raise NotImplementedError()




import datetime as dt
from collections import namedtuple

import pandas as pd

from .lib import inplace_decay
from ..lib.date import round_hour
from ..stoats.names import like, _d, FITNESS_D_ANY, FATIGUE_D_ANY

IMPULSE_3600 = 'Impulse / 3600s'
DecayModel = namedtuple('DecayModel', 'start, zero, log10_scale, log10_period, input, output')


def pre_calc(source, model, start=None, finish=None, target=None):
    # sum into 1-hour blocks (the decay is much longer period so this has little effect on accuracy
    # but saves time) and then add target indices
    # target here is intended for when fitting to data - that's the target data being fitted
    # (so we generate data at the right times)
    if target is not None:
        if start:
            start = min(start, target.index[0])
        else:
            start = target.index[0]
        if finish:
            finish = max(finish, target.index[-1])
        else:
            finish = target.index[-1]
    else:
        start = source.index[0]
        finish = source.index[-1]
    start, finish = round_hour(start, up=False), round_hour(finish, up=True)
    data = source.resample('1h', label='right').sum()
    data = data.loc[:, [model.input]]
    data.rename(columns={model.input: IMPULSE_3600}, inplace=True)
    times = pd.DataFrame(index=pd.date_range(start=start, end=finish, freq='1H'), columns=[model.output])
    data = data.join(times, how='outer', sort=True)
    t0 = data.index[0]
    initial = pd.DataFrame({IMPULSE_3600: model.start, model.output: 0}, index=[t0 - dt.timedelta(hours=1)])
    data = initial.append(data, sort=True)
    data.fillna(0, inplace=True)
    return data


def calc(data, model):
    data[model.output] = model.zero + data[IMPULSE_3600] * 10 ** model.log10_scale
    inplace_decay(data, model.output, 10 ** model.log10_period)
    return data


def impulse_stats(df):
    stats = {}
    for pattern in FITNESS_D_ANY, FATIGUE_D_ANY:
        for name in like(pattern, df.columns):
            stats[_d(name)] = df[name][-1] - df[name][0]
    return stats




from collections import namedtuple
from json import loads
from logging import getLogger
from re import split

import pandas as pd
import numpy as np

from ch2.data import present, linear_resample_time
from . import DataFrameCalculatorMixin, ActivityJournalCalculatorMixin, MultiProcCalculator
from ..load import StatisticJournalLoader
from ..names import *
from ...data import activity_statistics
from ...data.lib import interpolate_to_index
from ...data.power import add_differentials, add_energy_budget, add_loss_estimate, \
    add_power_estimate, PowerException, evaluate, fit_power, PowerModel, add_air_speed, add_modeled_hr
from ...lib.data import reftuple, MissingReference
from ...lib.log import log_current_exception
from ...squeal import StatisticJournalFloat, Constant, Timestamp

log = getLogger(__name__)
# these configure the model.
Power = reftuple('Power', 'bike, rider_weight, vary', defaults=(70, 'wind_speed, wind_heading, slope'))
Bike = namedtuple('Bike', 'cda, crr, weight')


# used as common owner
class PowerCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, **kargs):
        super().__init__(*args, owner_out=PowerCalculator, **kargs)
        self.power = None


class BasicPowerCalculator(PowerCalculator):

    def __init__(self, *args, power=None, caloric_eff=0.25, **kargs):
        self.power_ref = power
        self.caloric_eff = caloric_eff
        super().__init__(*args, **kargs)

    def _set_power(self, s, ajournal):
        power = Power(**loads(Constant.get(s, self.power_ref).at(s).value))
        # default owner is constant since that's what users can tweak
        self.power = power.expand(log, s, ajournal.start, owner=Constant, constraint=ajournal.activity_group)
        log.debug(f'Power: {self.power_ref}: {self.power}')

    def _read_dataframe(self, s, ajournal):
        try:
            self._set_power(s, ajournal)
            df = activity_statistics(s, DISTANCE, ELEVATION, SPEED, CADENCE, LATITUDE, LONGITUDE, HEART_RATE,
                                     activity_journal=ajournal, with_timespan=True)
            ldf = linear_resample_time(df)
            ldf = add_differentials(ldf)
            return df, ldf
        except PowerException as e:
            log.warning(e)
        except MissingReference as e:
            log.warning(f'Power configuration incorrect ({e})')
        except Exception as e:
            log.warning(f'Failed to generate statistics for power: {e}')
            raise

    def _calculate_stats(self, s, ajournal, dfs):
        df, ldf = dfs
        ldf = add_energy_budget(ldf, self.power.bike['weight'] + self.power.rider_weight)
        ldf = add_air_speed(ldf, 0, 0)
        ldf = add_loss_estimate(ldf, self.power.bike['cda'], self.power.bike['crr'])
        ldf = add_power_estimate(ldf)
        return df, ldf

    def _copy_results(self, s, ajournal, loader, dfs,
                      fields=((POWER_ESTIMATE, W, AVG), (HEADING, DEG, None))):
        df, ldf = dfs
        self.__add_total_energy(s, ajournal, loader, ldf)
        df = interpolate_to_index(df, ldf, *(field[0] for field in fields))
        for time, row in df.iterrows():
            for name, units, summary in fields:
                if not pd.isnull(row[name]):
                    loader.add(name, units, summary, ajournal.activity_group, ajournal, row[name], time,
                               StatisticJournalFloat)

    def __add_total_energy(self, s, ajournal, loader, ldf):
        if present(ldf, POWER_ESTIMATE):
            ldf['tmp'] = ldf[POWER_ESTIMATE]
            ldf.loc[ldf['tmp'].isna(), ['tmp']] = 0
            energy = np.trapz(y=ldf['tmp'], x=ldf.index.astype(np.int64) / 1e12)
            loader.add(ENERGY_ESTIMATE, KJ, MAX, ajournal.activity_group, ajournal, energy, ajournal.start,
                       StatisticJournalFloat)
            loader.add(CALORIE_ESTIMATE, KCAL, MAX, ajournal.activity_group, ajournal,
                       energy * 0.239006 / self.caloric_eff, ajournal.start, StatisticJournalFloat)
            ldf.drop(columns=['tmp'], inplace=True)


class ExtendedPowerCalculator(BasicPowerCalculator):
    '''
    This was an experiment that honestly didn't work too well.

    The idea was to extend the power model with constant speed / heading wind, and to fit for that,
    so that we included a wind correction.  To constrain the fitting I took the power, scaled and lagged,
    and compared it to heart rate.  The best scale/lag and wind model gave the power estimate,

    However, in testing, none of the windiest routes were the windy days cycling back down the Maipo valley.
    Instead, it tended to pick interval training when riding a loop.  Obviously fitting the pattern in the
    activity, not the wind.

    And it increased loading times so much it drove the re-implementation with multiple processes.
    '''

    # lots of fitting
    def __init__(self, *args, cost_calc=100, **kargs):
        super().__init__(*args, cost_calc=cost_calc, **kargs)

    def _run_one(self, s, time_or_date):
        source = self._get_source(s, time_or_date)
        s.commit()  # free up database
        with Timestamp(owner=self.owner_out, key=source.id).on_success(log, s):
            try:
                data = self._read_dataframe(s, source)
                loader = StatisticJournalLoader(s, self.owner_out)
                try:
                    stats = self._calculate_stats(s, source, data)
                except PowerException as e:
                    log.warning(f'Cannot use detailed power model; adding basic values only ({e})')
                    loader = StatisticJournalLoader(s, self.owner_out)
                    stats = None, super()._calculate_stats(s, source, data)
                self._copy_results(s, source, loader, stats)
                loader.load()
            except Exception as e:
                log.warning(f'No statistics on {time_or_date}')
                log_current_exception()

    def __varying(self):
        return list(filter(None, split(r'[\s,]*([^, ]+)[\s ]*', self.power.vary)))

    def _calculate_stats(self, s, ajournal, df):
        vary = self.__varying()
        if not vary:
            raise PowerException('No parameters to vary - fitting disabled')
        model = PowerModel(cda=self.power.bike['cda'], crr=self.power.bike['crr'],
                           m=self.power.bike['weight'] + self.power.rider_weight,
                           wind_speed=10, wind_heading=180)
        for name in self.power._fields:
            if name in model._fields:
                setattr(model, name, getattr(self.power, name))
        model = fit_power(df, model, *vary)
        df = evaluate(df, model, quiet=False)
        df = add_modeled_hr(df, model.window, model.slope, model.delay)
        # p_hr = 60 / model.slope
        # if p_hr < 100 or p_hr > 500:
        #     raise PowerException(f'Unreasonable model results (slope {model.slope} / {p_hr})')
        # if model.delay > 30:
        #     raise PowerException(f'Unreasonable model results (delay {model.delay})')
        return model, df

    def _copy_results(self, s, ajournal, loader, stats):
        model, df = stats
        fields = ((POWER_ESTIMATE, W, AVG), (HEADING, DEG, None))
        if model:
            # how much energy every heart beat
            # 60W at 60bpm is 60J every second or beat; 60W at 1bpm is 3600J every minute or beat;
            # 1W at 1bpm is 60J every minute or beat
            # slope is BPM / W; 1/slope is W/BPM = W/PM = WM = 60Ws
            vary = self.__varying()
            if 'slope' in vary:
                loader.add(POWER_HR, J, AVG, ajournal.activity_group, ajournal, 60 / model.slope, ajournal.start,
                           StatisticJournalFloat)
            if 'delay' in vary:
                loader.add(POWER_HR_LAG, S, AVG, ajournal.activity_group, ajournal, model.delay,
                           ajournal.start, StatisticJournalFloat)
            if 'wind_speed' in vary:
                loader.add(WIND_SPEED, MS, AVG, ajournal.activity_group, ajournal, model.wind_speed,
                           ajournal.start, StatisticJournalFloat)
            if 'wind_heading' in vary:
                loader.add(WIND_HEADING, DEG, AVG, ajournal.activity_group, ajournal, model.wind_heading,
                           ajournal.start, StatisticJournalFloat)
            fields = fields + ((PREDICTED_HEART_RATE, BPM, None), (DETRENDED_HEART_RATE, BPM, None))
        # has to come after the above to get times in order
        super()._copy_results(s, ajournal, loader, df, fields=fields)




import datetime as dt

from bokeh.io import output_file
from bokeh.layouts import column, row, gridplot
from bokeh.plotting import show

from ch2.data import *
from ch2.lib import *
from ch2.stoats.names import _log
from ch2.uranus.decorator import template


@template
def compare_activities(local_time: to_date, compare_time: to_date, activity_group_name):

    f'''
    # Compare Activities: {local_time.split()[0]} v {compare_time.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')

    activity = std_activity_statistics(s, local_time=local_time, activity_group_name=activity_group_name)
    compare = std_activity_statistics(s, local_time=compare_time, activity_group_name=activity_group_name)
    details = activity_statistics(s, 'Climb %', ACTIVE_TIME, ACTIVE_DISTANCE, local_time=local_time,
                                  activity_group_name=activity_group_name)
    health = std_health_statistics(s)

    f'''
    ## Activity Plots
    
    The black line shows data from {local_time.split()[0]}, 
    the grey line from {compare_time.split()[0]}. 
    To the right of each plot of data against distance is a related plot of cumulative data
    (except the last, cadence, which isn't useful and so replaced by HR zones).
    Green and red areas indicate differences between the two dates. 
    Additional red lines on the altitude plot are auto-detected climbs.
    
    Plot tools support zoom, dragging, etc.
    '''

    output_file(filename='/dev/null')

    el = comparison_line_plot(700, 200, DISTANCE_KM, ELEVATION_M, activity, other=compare)
    add_climbs(el, details, activity)
    el_c = cumulative_plot(200, 200, CLIMB_MS, activity, other=compare)

    sp = comparison_line_plot(700, 200, DISTANCE_KM, MED_SPEED_KMH, activity, other=compare, ylo=0, x_range=el.x_range)
    sp_c = cumulative_plot(200, 200, MED_SPEED_KMH, activity, other=compare, ylo=0)

    hri = comparison_line_plot(700, 200, DISTANCE_KM, HR_IMPULSE_10, activity, other=compare, ylo=0, x_range=el.x_range)
    hri_c = cumulative_plot(200, 200, HR_IMPULSE_10, activity, other=compare, ylo=0)

    hr = comparison_line_plot(700, 200, DISTANCE_KM, HEART_RATE_BPM, activity, other=compare, ylo=0, x_range=el.x_range)
    hr_c = cumulative_plot(200, 200, HEART_RATE_BPM, activity, other=compare, ylo=0)

    if present(activity, MED_POWER_ESTIMATE_W):
        pw = comparison_line_plot(700, 200, DISTANCE_KM, MED_POWER_ESTIMATE_W, activity, other=compare, ylo=0, x_range=el.x_range)
        pw_c = cumulative_plot(200, 200, MED_POWER_ESTIMATE_W, activity, other=compare, ylo=0)
    else:
        pw, pw_c = None, None

    if present(activity, MED_CADENCE):
        cd = comparison_line_plot(700, 200, DISTANCE_KM, MED_CADENCE, activity, other=compare, ylo=0, x_range=el.x_range)
    else:
        cd = None
    hr_h = histogram_plot(200, 200, HR_ZONE, activity, xlo=1, xhi=5)

    show(gridplot([[el, el_c], [sp, sp_c], [hri, hri_c], [hr, hr_c], [pw, pw_c], [cd, hr_h]]))

    '''
    ## Activity Maps
    '''

    map = map_plot(400, 400, activity, other=compare)
    m_el = map_intensity(200, 200, activity, ELEVATION_M, ranges=map)
    m_sp = map_intensity(200, 200, activity, SPEED_KMH, ranges=map)
    m_hr = map_intensity(200, 200, activity, HR_IMPULSE_10, ranges=map)
    if present(activity, MED_POWER_ESTIMATE_W):
        m_pw = map_intensity(200, 200, activity, MED_POWER_ESTIMATE_W, ranges=map)
    else:
        m_pw = None
    show(row(map, gridplot([[m_el, m_sp], [m_hr, m_pw]], toolbar_location='right')))

    '''
    ## Activity Statistics
    '''

    '''
    Active time and distance exclude pauses.
    '''

    details[[ACTIVE_TIME, ACTIVE_DISTANCE]].dropna(). \
        transform({ACTIVE_TIME: format_seconds, ACTIVE_DISTANCE: format_metres})

    '''
    Climbs are auto-detected and shown only for the main activity. They are included in the elevation plot above.
    '''

    details.filter(like='Climb').dropna(). \
        transform({CLIMB_TIME: format_seconds, CLIMB_ELEVATION: format_metres,
                   CLIMB_DISTANCE: format_metres, CLIMB_GRADIENT: format_percent})

    '''
    ## Health and Fitness
    '''

    fitness, fatigue = like(FITNESS_D_ANY, health.columns), like(FATIGUE_D_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, TIME, fitness + fatigue, health, colours, alphas=alphas)
    log_ff = multi_line_plot(900, 100, TIME, [_log(name) for name in fitness + fatigue], health, colours,
                             alphas=alphas, x_range=ff.x_range, y_label='Log FF')
    atd = multi_dot_plot(900, 200, TIME, [ACTIVE_TIME_H, ACTIVE_DISTANCE_KM], health, ['black', 'grey'], alphas=[1, 0.5],
                         x_range=ff.x_range, rescale=True)
    shr = multi_plot(900, 200, TIME, [DAILY_STEPS, REST_HR], health, ['grey', 'red'], alphas=[1, 0.5],
                     x_range=ff.x_range, rescale=True, plotters=[bar_plotter(dt.timedelta(hours=20)), dot_plotter()])
    show(column(ff, log_ff, atd, shr))




# lets us define a template in python that IntelliJ will validate without flagging an unused code error.
TEMPLATE = lambda *args: False

# need to import these so they are available for introspection
from . import activity_details, all_activities, calendar, health, similar_activities, fit_ff_parameters